import itertools
import logging
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Dict, List, Callable, Tuple, Any

import pandas as pd

from autorag.schema.module import Module
from autorag.support import get_support_nodes
from autorag.utils.util import make_combinations, explode, find_key_values

logger = logging.getLogger("AutoRAG")


@dataclass
class Node:
	node_type: str
	strategy: Dict
	node_params: Dict
	modules: List[Module]
	run_node: Callable = field(init=False)

	def __post_init__(self):
		self.run_node = get_support_nodes(self.node_type)
		if self.run_node is None:
			raise ValueError(f"Node type {self.node_type} is not supported.")

	def get_param_combinations(self) -> Tuple[List[Callable], List[Dict]]:
		"""
		This method returns a combination of module and node parameters, also corresponding modules.

		:return: Each module and its module parameters.
		:rtype: Tuple[List[Callable], List[Dict]]
		"""

		def make_single_combination(module: Module) -> List[Dict]:
			input_dict = {**self.node_params, **module.module_param}
			return make_combinations(input_dict)

		combinations = list(map(make_single_combination, self.modules))
		module_list, combination_list = explode(self.modules, combinations)
		return list(map(lambda x: x.module, module_list)), combination_list

	@classmethod
	def from_dict(cls, node_dict: Dict) -> "Node":
		_node_dict = deepcopy(node_dict)
		node_type = _node_dict.pop("node_type")
		strategy = _node_dict.pop("strategy")
		modules = list(map(lambda x: Module.from_dict(x), _node_dict.pop("modules")))
		node_params = _node_dict
		return cls(node_type, strategy, node_params, modules)

	def run(self, previous_result: pd.DataFrame, node_line_dir: str) -> pd.DataFrame:
		logger.info(f"Running node {self.node_type}...")
		input_modules, input_params = self.get_param_combinations()
		return self.run_node(
			modules=input_modules,
			module_params=input_params,
			previous_result=previous_result,
			node_line_dir=node_line_dir,
			strategies=self.strategy,
		)


def extract_values(node: Node, key: str) -> List[str]:
	"""
	This function extract values from node's modules' module_param.

	:param node: The node you want to extract values from.
	:param key: The key of module_param that you want to extract.
	:return: The list of extracted values.
	    It removes duplicated elements automatically.
	"""

	def extract_module_values(module: Module):
		if key not in module.module_param:
			return []
		value = module.module_param[key]
		if isinstance(value, str) or isinstance(value, int):
			return [value]
		elif isinstance(value, list):
			return value
		else:
			raise ValueError(f"{key} must be str,list or int, but got {type(value)}")

	values = list(map(extract_module_values, node.modules))
	return list(set(list(itertools.chain.from_iterable(values))))


def extract_values_from_nodes(nodes: List[Node], key: str) -> List[str]:
	"""
	This function extract values from nodes' modules' module_param.

	:param nodes: The nodes you want to extract values from.
	:param key: The key of module_param that you want to extract.
	:return: The list of extracted values.
	    It removes duplicated elements automatically.
	"""
	values = list(map(lambda node: extract_values(node, key), nodes))
	return list(set(list(itertools.chain.from_iterable(values))))


def extract_values_from_nodes_strategy(nodes: List[Node], key: str) -> List[Any]:
	"""
	This function extract values from nodes' strategy.

	:param nodes: The nodes you want to extract values from.
	:param key: The key string that you want to extract.
	:return: The list of extracted values.
	    It removes duplicated elements automatically.
	"""
	values = []
	for node in nodes:
		value_list = find_key_values(node.strategy, key)
		if value_list:
			values.extend(value_list)
	return values


def module_type_exists(nodes: List[Node], module_type: str) -> bool:
	"""
	This function check if the module type exists in the nodes.

	:param nodes: The nodes you want to check.
	:param module_type: The module type you want to check.
	:return: True if the module type exists in the nodes.
	"""
	return any(
		list(
			map(
				lambda node: any(
					list(
						map(
							lambda module: module.module_type.lower()
							== module_type.lower(),
							node.modules,
						)
					)
				),
				nodes,
			)
		)
	)



import ast
import asyncio
import datetime
import functools
import glob
import inspect
import itertools
import json
import logging
import os
import re
import string
from copy import deepcopy
from json import JSONDecoder
from typing import List, Callable, Dict, Optional, Any, Collection, Iterable, Union

from asyncio import AbstractEventLoop
import emoji
import numpy as np
import pandas as pd
import tiktoken
import unicodedata

import yaml
from llama_index.embeddings.openai import OpenAIEmbedding
from pydantic import BaseModel as BM
from pydantic.v1 import BaseModel

logger = logging.getLogger("AutoRAG")


def fetch_contents(
	corpus_data: pd.DataFrame, ids: List[List[str]], column_name: str = "contents"
) -> List[List[Any]]:
	def fetch_contents_pure(
		ids: List[str], corpus_data: pd.DataFrame, column_name: str
	):
		return list(map(lambda x: fetch_one_content(corpus_data, x, column_name), ids))

	result = flatten_apply(
		fetch_contents_pure, ids, corpus_data=corpus_data, column_name=column_name
	)
	return result


def fetch_one_content(
	corpus_data: pd.DataFrame,
	id_: str,
	column_name: str = "contents",
	id_column_name: str = "doc_id",
) -> Any:
	if isinstance(id_, str):
		if id_ in ["", ""]:
			return None
		fetch_result = corpus_data[corpus_data[id_column_name] == id_]
		if fetch_result.empty:
			raise ValueError(f"doc_id: {id_} not found in corpus_data.")
		else:
			return fetch_result[column_name].iloc[0]
	else:
		return None


def result_to_dataframe(column_names: List[str]):
	"""
	Decorator for converting results to pd.DataFrame.
	"""

	def decorator_result_to_dataframe(func: Callable):
		@functools.wraps(func)
		def wrapper(*args, **kwargs) -> pd.DataFrame:
			results = func(*args, **kwargs)
			if len(column_names) == 1:
				df_input = {column_names[0]: results}
			else:
				df_input = {
					column_name: result
					for result, column_name in zip(results, column_names)
				}
			result_df = pd.DataFrame(df_input)
			return result_df

		return wrapper

	return decorator_result_to_dataframe


def load_summary_file(
	summary_path: str, dict_columns: Optional[List[str]] = None
) -> pd.DataFrame:
	"""
	Load a summary file from summary_path.

	:param summary_path: The path of the summary file.
	:param dict_columns: The columns that are dictionary type.
	    You must fill this parameter if you want to load summary file properly.
	    Default is ['module_params'].
	:return: The summary dataframe.
	"""
	if not os.path.exists(summary_path):
		raise ValueError(f"summary.csv does not exist in {summary_path}.")
	summary_df = pd.read_csv(summary_path)
	if dict_columns is None:
		dict_columns = ["module_params"]

	if any([col not in summary_df.columns for col in dict_columns]):
		raise ValueError(f"{dict_columns} must be in summary_df.columns.")

	def convert_dict(elem):
		try:
			return ast.literal_eval(elem)
		except:
			# convert datetime or date to its object (recency filter)
			date_object = convert_datetime_string(elem)
			if date_object is None:
				raise ValueError(
					f"Malformed dict received : {elem}\nCan't convert to dict properly"
				)
			return {"threshold": date_object}

	summary_df[dict_columns] = summary_df[dict_columns].map(convert_dict)
	return summary_df


def convert_datetime_string(s):
	# Regex to extract datetime arguments from the string
	m = re.search(r"(datetime|date)(\((\d+)(,\s*\d+)*\))", s)
	if m:
		args = ast.literal_eval(m.group(2))
		if m.group(1) == "datetime":
			return datetime.datetime(*args)
		elif m.group(1) == "date":
			return datetime.date(*args)
	return None


def make_combinations(target_dict: Dict[str, Any]) -> List[Dict[str, Any]]:
	"""
	Make combinations from target_dict.
	The target_dict key value must be a string,
	and the value can be a list of values or single value.
	If generates all combinations of values from target_dict,
	which means generating dictionaries that contain only one value for each key,
	and all dictionaries will be different from each other.

	:param target_dict: The target dictionary.
	:return: The list of generated dictionaries.
	"""
	dict_with_lists = dict(
		map(
			lambda x: (x[0], x[1] if isinstance(x[1], list) else [x[1]]),
			target_dict.items(),
		)
	)

	def delete_duplicate(x):
		def is_hashable(obj):
			try:
				hash(obj)
				return True
			except TypeError:
				return False

		if any([not is_hashable(elem) for elem in x]):
			# TODO: add duplication check for unhashable objects
			return x
		else:
			return list(set(x))

	dict_with_lists = dict(
		map(lambda x: (x[0], delete_duplicate(x[1])), dict_with_lists.items())
	)
	combination = list(itertools.product(*dict_with_lists.values()))
	combination_dicts = [
		dict(zip(dict_with_lists.keys(), combo)) for combo in combination
	]
	return combination_dicts


def explode(index_values: Collection[Any], explode_values: Collection[Collection[Any]]):
	"""
	Explode index_values and explode_values.
	The index_values and explode_values must have the same length.
	It will flatten explode_values and keep index_values as a pair.

	:param index_values: The index values.
	:param explode_values: The exploded values.
	:return: Tuple of exploded index_values and exploded explode_values.
	"""
	assert len(index_values) == len(
		explode_values
	), "Index values and explode values must have same length"
	df = pd.DataFrame({"index_values": index_values, "explode_values": explode_values})
	df = df.explode("explode_values")
	return df["index_values"].tolist(), df["explode_values"].tolist()


def replace_value_in_dict(target_dict: Dict, key: str, replace_value: Any) -> Dict:
	"""
	Replace the value of a certain key in target_dict.
	If there is no targeted key in target_dict, it will return target_dict.

	:param target_dict: The target dictionary.
	:param key: The key is to replace.
	:param replace_value: The value to replace.
	:return: The replaced dictionary.
	"""
	replaced_dict = deepcopy(target_dict)
	if key not in replaced_dict:
		return replaced_dict
	replaced_dict[key] = replace_value
	return replaced_dict


def normalize_string(s: str) -> str:
	"""
	Taken from the official evaluation script for v1.1 of the SQuAD dataset.
	Lower text and remove punctuation, articles, and extra whitespace.
	"""

	def remove_articles(text):
		return re.sub(r"\b(a|an|the)\b", " ", text)

	def white_space_fix(text):
		return " ".join(text.split())

	def remove_punc(text):
		exclude = set(string.punctuation)
		return "".join(ch for ch in text if ch not in exclude)

	def lower(text):
		return text.lower()

	return white_space_fix(remove_articles(remove_punc(lower(s))))


def convert_string_to_tuple_in_dict(d):
	"""Recursively converts strings that start with '(' and end with ')' to tuples in a dictionary."""
	for key, value in d.items():
		# If the value is a dictionary, recurse
		if isinstance(value, dict):
			convert_string_to_tuple_in_dict(value)
		# If the value is a list, iterate through its elements
		elif isinstance(value, list):
			for i, item in enumerate(value):
				# If an item in the list is a dictionary, recurse
				if isinstance(item, dict):
					convert_string_to_tuple_in_dict(item)
				# If an item in the list is a string matching the criteria, convert it to a tuple
				elif (
					isinstance(item, str)
					and item.startswith("(")
					and item.endswith(")")
				):
					value[i] = ast.literal_eval(item)
		# If the value is a string matching the criteria, convert it to a tuple
		elif isinstance(value, str) and value.startswith("(") and value.endswith(")"):
			d[key] = ast.literal_eval(value)

	return d


def convert_env_in_dict(d: Dict):
	"""
	Recursively converts environment variable string in a dictionary to actual environment variable.

	:param d: The dictionary to convert.
	:return: The converted dictionary.
	"""
	env_pattern = re.compile(r".*?\${(.*?)}.*?")

	def convert_env(val: str):
		matches = env_pattern.findall(val)
		for match in matches:
			val = val.replace(f"${{{match}}}", os.environ.get(match, ""))
		return val

	for key, value in d.items():
		if isinstance(value, dict):
			convert_env_in_dict(value)
		elif isinstance(value, list):
			for i, item in enumerate(value):
				if isinstance(item, dict):
					convert_env_in_dict(item)
				elif isinstance(item, str):
					value[i] = convert_env(item)
		elif isinstance(value, str):
			d[key] = convert_env(value)
	return d


async def process_batch(tasks, batch_size: int = 64) -> List[Any]:
	"""
	Processes tasks in batches asynchronously.

	:param tasks: A list of no-argument functions or coroutines to be executed.
	:param batch_size: The number of tasks to process in a single batch.
	    Default is 64.
	:return: A list of results from the processed tasks.
	"""
	results = []

	for i in range(0, len(tasks), batch_size):
		batch = tasks[i : i + batch_size]
		batch_results = await asyncio.gather(*batch)
		results.extend(batch_results)

	return results


def make_batch(elems: List[Any], batch_size: int) -> List[List[Any]]:
	"""
	Make a batch of elems with batch_size.
	"""
	return [elems[i : i + batch_size] for i in range(0, len(elems), batch_size)]


def save_parquet_safe(df: pd.DataFrame, filepath: str, upsert: bool = False):
	output_file_dir = os.path.dirname(filepath)
	if not os.path.isdir(output_file_dir):
		raise NotADirectoryError(f"directory {output_file_dir} not found.")
	if not filepath.endswith("parquet"):
		raise NameError(
			f'file path: {filepath}  filename extension need to be ".parquet"'
		)
	if os.path.exists(filepath) and not upsert:
		raise FileExistsError(
			f"file {filepath} already exists."
			"Set upsert True if you want to overwrite the file."
		)

	df.to_parquet(filepath, index=False)


def openai_truncate_by_token(
	texts: List[str], token_limit: int, model_name: str
) -> List[str]:
	try:
		tokenizer = tiktoken.encoding_for_model(model_name)
	except KeyError:
		# This is not a real OpenAI model
		return texts

	def truncate_text(text: str, limit: int, tokenizer):
		tokens = tokenizer.encode(text)
		if len(tokens) <= limit:
			return text
		truncated_text = tokenizer.decode(tokens[:limit])
		return truncated_text

	return list(map(lambda x: truncate_text(x, token_limit, tokenizer), texts))


def reconstruct_list(flat_list: List[Any], lengths: List[int]) -> List[List[Any]]:
	result = []
	start = 0
	for length in lengths:
		result.append(flat_list[start : start + length])
		start += length
	return result


def flatten_apply(
	func: Callable, nested_list: List[List[Any]], **kwargs
) -> List[List[Any]]:
	"""
	This function flattens the input list and applies the function to the elements.
	After that, it reconstructs the list to the original shape.
	Its speciality is that the first dimension length of the list can be different from each other.

	:param func: The function that applies to the flattened list.
	:param nested_list: The nested list to be flattened.
	:return: The list that is reconstructed after applying the function.
	"""
	df = pd.DataFrame({"col1": nested_list})
	df = df.explode("col1")
	df["result"] = func(df["col1"].tolist(), **kwargs)
	return df.groupby(level=0, sort=False)["result"].apply(list).tolist()


async def aflatten_apply(
	func: Callable, nested_list: List[List[Any]], **kwargs
) -> List[List[Any]]:
	"""
	This function flattens the input list and applies the function to the elements.
	After that, it reconstructs the list to the original shape.
	Its speciality is that the first dimension length of the list can be different from each other.

	:param func: The function that applies to the flattened list.
	:param nested_list: The nested list to be flattened.
	:return: The list that is reconstructed after applying the function.
	"""
	df = pd.DataFrame({"col1": nested_list})
	df = df.explode("col1")
	df["result"] = await func(df["col1"].tolist(), **kwargs)
	return df.groupby(level=0, sort=False)["result"].apply(list).tolist()


def sort_by_scores(row, reverse=True):
	"""
	Sorts each row by 'scores' column.
	The input column names must be 'contents', 'ids', and 'scores'.
	And its elements must be list type.
	"""
	results = sorted(
		zip(row["contents"], row["ids"], row["scores"]),
		key=lambda x: x[2],
		reverse=reverse,
	)
	reranked_contents, reranked_ids, reranked_scores = zip(*results)
	return list(reranked_contents), list(reranked_ids), list(reranked_scores)


def select_top_k(df, column_names: List[str], top_k: int):
	for column_name in column_names:
		df[column_name] = df[column_name].apply(lambda x: x[:top_k])
	return df


def filter_dict_keys(dict_, keys: List[str]):
	result = {}
	for key in keys:
		if key in dict_:
			result[key] = dict_[key]
		else:
			raise KeyError(f"Key '{key}' not found in dictionary.")
	return result


def split_dataframe(df, chunk_size):
	num_chunks = (
		len(df) // chunk_size + 1
		if len(df) % chunk_size != 0
		else len(df) // chunk_size
	)
	result = list(
		map(lambda x: df[x * chunk_size : (x + 1) * chunk_size], range(num_chunks))
	)
	result = list(map(lambda x: x.reset_index(drop=True), result))
	return result


def find_trial_dir(project_dir: str) -> List[str]:
	# Pattern to match directories named with numbers
	pattern = os.path.join(project_dir, "[0-9]*")
	all_entries = glob.glob(pattern)

	# Filter out only directories
	trial_dirs = [
		entry
		for entry in all_entries
		if os.path.isdir(entry) and entry.split(os.sep)[-1].isdigit()
	]

	return trial_dirs


def find_node_summary_files(trial_dir: str) -> List[str]:
	# Find all summary.csv files recursively
	all_summary_files = glob.glob(
		os.path.join(trial_dir, "**", "summary.csv"), recursive=True
	)

	# Filter out files that are at a lower directory level
	filtered_files = [
		f for f in all_summary_files if f.count(os.sep) > trial_dir.count(os.sep) + 2
	]

	return filtered_files


def preprocess_text(text: str) -> str:
	return normalize_unicode(demojize(text))


def demojize(text: str) -> str:
	return emoji.demojize(text)


def normalize_unicode(text: str) -> str:
	return unicodedata.normalize("NFC", text)


def dict_to_markdown(d, level=1):
	"""
	Convert a dictionary to a Markdown formatted string.

	:param d: Dictionary to convert
	:param level: Current level of heading (used for nested dictionaries)
	:return: Markdown formatted string
	"""
	markdown = ""
	for key, value in d.items():
		if isinstance(value, dict):
			markdown += f"{'#' * level} {key}\n"
			markdown += dict_to_markdown(value, level + 1)
		elif isinstance(value, list):
			markdown += f"{'#' * level} {key}\n"
			for item in value:
				if isinstance(item, dict):
					markdown += dict_to_markdown(item, level + 1)
				else:
					markdown += f"- {item}\n"
		else:
			markdown += f"{'#' * level} {key}\n{value}\n"
	return markdown


def dict_to_markdown_table(data, key_column_name: str, value_column_name: str):
	# Check if the input is a dictionary
	if not isinstance(data, dict):
		raise ValueError("Input must be a dictionary")

	# Create the header of the table
	header = f"| {key_column_name} | {value_column_name} |\n| :---: | :-----: |\n"

	# Create the rows of the table
	rows = ""
	for key, value in data.items():
		rows += f"| {key} | {value} |\n"

	# Combine header and rows
	markdown_table = header + rows
	return markdown_table


def embedding_query_content(
	queries: List[str],
	contents_list: List[List[str]],
	embedding_model: Optional[str] = None,
	batch: int = 128,
):
	flatten_contents = list(itertools.chain.from_iterable(contents_list))

	openai_embedding_limit = 8000  # all openai embedding model has 8000 max token input
	if isinstance(embedding_model, OpenAIEmbedding):
		queries = openai_truncate_by_token(
			queries, openai_embedding_limit, embedding_model.model_name
		)
		flatten_contents = openai_truncate_by_token(
			flatten_contents, openai_embedding_limit, embedding_model.model_name
		)

	# Embedding using batch
	embedding_model.embed_batch_size = batch
	query_embeddings = embedding_model.get_text_embedding_batch(queries)

	content_lengths = list(map(len, contents_list))
	content_embeddings_flatten = embedding_model.get_text_embedding_batch(
		flatten_contents
	)
	content_embeddings = reconstruct_list(content_embeddings_flatten, content_lengths)
	return query_embeddings, content_embeddings


def to_list(item):
	"""Recursively convert collections to Python lists."""
	if isinstance(item, np.ndarray):
		# Convert numpy array to list and recursively process each element
		return [to_list(sub_item) for sub_item in item.tolist()]
	elif isinstance(item, pd.Series):
		# Convert pandas Series to list and recursively process each element
		return [to_list(sub_item) for sub_item in item.tolist()]
	elif isinstance(item, Iterable) and not isinstance(
		item, (str, bytes, BaseModel, BM)
	):
		# Recursively process each element in other iterables
		return [to_list(sub_item) for sub_item in item]
	else:
		return item


def convert_inputs_to_list(func):
	"""Decorator to convert all function inputs to Python lists."""

	@functools.wraps(func)
	def wrapper(*args, **kwargs):
		new_args = [to_list(arg) for arg in args]
		new_kwargs = {k: to_list(v) for k, v in kwargs.items()}
		return func(*new_args, **new_kwargs)

	return wrapper


def get_best_row(
	summary_df: pd.DataFrame, best_column_name: str = "is_best"
) -> pd.Series:
	"""
	From the summary dataframe, find the best result row by 'is_best' column and return it.

	:param summary_df: Summary dataframe created by AutoRAG.
	:param best_column_name: The column name that indicates the best result.
	    Default is 'is_best'.
	    You don't have to change this unless the column name is different.
	:return: Best row pandas Series instance.
	"""
	bests = summary_df.loc[summary_df[best_column_name]]
	assert len(bests) == 1, "There must be only one best result."
	return bests.iloc[0]


def get_event_loop() -> AbstractEventLoop:
	"""
	Get asyncio event loop safely.
	"""
	try:
		loop = asyncio.get_running_loop()
	except RuntimeError:
		loop = asyncio.new_event_loop()
		asyncio.set_event_loop(loop)
	return loop


def find_key_values(data, target_key: str) -> List[Any]:
	"""
	Recursively find all values for a specific key in a nested dictionary or list.

	:param data: The dictionary or list to search.
	:param target_key: The key to search for.
	:return: A list of values associated with the target key.
	"""
	values = []

	if isinstance(data, dict):
		for key, value in data.items():
			if key == target_key:
				values.append(value)
			if isinstance(value, (dict, list)):
				values.extend(find_key_values(value, target_key))
	elif isinstance(data, list):
		for item in data:
			if isinstance(item, (dict, list)):
				values.extend(find_key_values(item, target_key))

	return values


def pop_params(func: Callable, kwargs: Dict) -> Dict:
	"""
	Pop parameters from the given func and return them.
	It automatically deletes the parameters like "self" or "cls".

	:param func: The function to pop parameters.
	:param kwargs: kwargs to pop parameters.
	:return: The popped parameters.
	"""
	ignore_params = ["self", "cls"]
	target_params = list(inspect.signature(func).parameters.keys())
	target_params = list(filter(lambda x: x not in ignore_params, target_params))

	init_params = {}
	kwargs_keys = list(kwargs.keys())
	for key in kwargs_keys:
		if key in target_params:
			init_params[key] = kwargs.pop(key)
	return init_params


def apply_recursive(func, data):
	"""
	Recursively apply a function to all elements in a list, tuple, set, np.ndarray, or pd.Series and return as List.

	:param func: Function to apply to each element.
	:param data: List or nested list.
	:return: List with the function applied to each element.
	"""
	if (
		isinstance(data, list)
		or isinstance(data, tuple)
		or isinstance(data, set)
		or isinstance(data, np.ndarray)
		or isinstance(data, pd.Series)
	):
		return [apply_recursive(func, item) for item in data]
	else:
		return func(data)


def empty_cuda_cache():
	try:
		import torch

		if torch.cuda.is_available():
			torch.cuda.empty_cache()
	except ImportError:
		pass


def load_yaml_config(yaml_path: str) -> Dict:
	"""
	Load a YAML configuration file for AutoRAG.
	It contains safe loading, converting string to tuple, and insert environment variables.

	:param yaml_path: The path of the YAML configuration file.
	:return: The loaded configuration dictionary.
	"""
	if not os.path.exists(yaml_path):
		raise ValueError(f"YAML file {yaml_path} does not exist.")
	with open(yaml_path, "r", encoding="utf-8") as stream:
		try:
			yaml_dict = yaml.safe_load(stream)
		except yaml.YAMLError as exc:
			raise ValueError(f"YAML file {yaml_path} could not be loaded.") from exc

	yaml_dict = convert_string_to_tuple_in_dict(yaml_dict)
	yaml_dict = convert_env_in_dict(yaml_dict)
	return yaml_dict


def decode_multiple_json_from_bytes(byte_data: bytes) -> list:
	"""
	Decode multiple JSON objects from bytes received from SSE server.

	Args:
	        byte_data: Bytes containing one or more JSON objects

	Returns:
	        List of decoded JSON objects
	"""
	# Decode bytes to string
	try:
		text_data = byte_data.decode("utf-8").strip()
	except UnicodeDecodeError:
		raise ValueError("Invalid byte data: Unable to decode as UTF-8")

	# Initialize decoder and result list
	decoder = JSONDecoder()
	result = []

	# Keep track of position in string
	pos = 0
	text_data = text_data.strip()

	while pos < len(text_data):
		try:
			# Try to decode next JSON object
			json_obj, json_end = decoder.raw_decode(text_data[pos:])
			result.append(json_obj)

			# Move position to end of current JSON object
			pos += json_end

			# Skip any whitespace
			while pos < len(text_data) and text_data[pos].isspace():
				pos += 1

		except json.JSONDecodeError:
			# If we can't decode at current position, move forward one character
			pos += 1

	return result


def is_chat_prompt(prompt: Union[List[str], List[List[dict]]]):
	if isinstance(prompt[0], str):
		return False
	if isinstance(prompt[0], list) and isinstance(prompt[0][0], dict):
		return True
	raise ValueError("prompt must be List[str] or List[List[dict]]")



import logging

from pinecone.grpc import PineconeGRPC as Pinecone_client
from pinecone import ServerlessSpec

from typing import List, Optional, Tuple, Union

from autorag.utils.util import make_batch, apply_recursive
from autorag.vectordb import BaseVectorStore

logger = logging.getLogger("AutoRAG")


class Pinecone(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		index_name: str,
		embedding_batch: int = 100,
		dimension: int = 1536,
		similarity_metric: str = "cosine",  # "cosine", "dotproduct", "euclidean"
		cloud: Optional[str] = "aws",
		region: Optional[str] = "us-east-1",
		api_key: Optional[str] = None,
		deletion_protection: Optional[str] = "disabled",  # "enabled" or "disabled"
		namespace: Optional[str] = "default",
		ingest_batch: int = 200,
	):
		super().__init__(embedding_model, similarity_metric, embedding_batch)

		self.index_name = index_name
		self.namespace = namespace
		self.ingest_batch = ingest_batch

		self.client = Pinecone_client(api_key=api_key)

		if similarity_metric == "ip":
			similarity_metric = "dotproduct"
		elif similarity_metric == "l2":
			similarity_metric = "euclidean"

		if not self.client.has_index(index_name):
			self.client.create_index(
				name=index_name,
				dimension=dimension,
				metric=similarity_metric,
				spec=ServerlessSpec(
					cloud=cloud,
					region=region,
				),
				deletion_protection=deletion_protection,
			)
		self.index = self.client.Index(index_name)

	async def add(self, ids: List[str], texts: List[str]):
		texts = self.truncated_inputs(texts)
		text_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(texts)
		self.add_embedding(ids, text_embeddings)

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		vector_tuples = list(zip(ids, embeddings))
		batch_vectors = make_batch(vector_tuples, self.ingest_batch)

		async_res = [
			self.index.upsert(
				vectors=batch_vector_tuples,
				namespace=self.namespace,
				async_req=True,
			)
			for batch_vector_tuples in batch_vectors
		]
		# Wait for the async requests to finish
		[async_result.result() for async_result in async_res]

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		results = self.index.fetch(ids=ids, namespace=self.namespace)
		id_vector_dict = {
			str(key): val["values"] for key, val in results["vectors"].items()
		}
		result = [id_vector_dict[_id] for _id in ids]
		return result

	async def is_exist(self, ids: List[str]) -> List[bool]:
		fetched_result = self.index.fetch(ids=ids, namespace=self.namespace)
		existed_ids = list(map(str, fetched_result.get("vectors", {}).keys()))
		return list(map(lambda x: x in existed_ids, ids))

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)

		ids, scores = [], []
		for query_embedding in query_embeddings:
			response = self.index.query(
				vector=query_embedding,
				top_k=top_k,
				include_values=True,
				namespace=self.namespace,
			)

			ids.append([o.id for o in response.matches])
			scores.append([o.score for o in response.matches])

		if self.similarity_metric in ["l2"]:
			scores = apply_recursive(lambda x: -x, scores)

		return ids, scores

	async def delete(self, ids: List[str]):
		# Delete entries by IDs
		self.index.delete(ids=ids, namespace=self.namespace)

	def delete_index(self):
		# Delete the index
		self.client.delete_index(self.index_name)



from itertools import chain
from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.nodes.passagereranker.tart.modeling_enc_t5 import (
	EncT5ForSequenceClassification,
)
from autorag.nodes.passagereranker.tart.tokenization_enc_t5 import EncT5Tokenizer
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	empty_cuda_cache,
)


class Tart(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir)
		try:
			import torch
		except ImportError:
			raise ImportError(
				"torch is not installed. Please install torch first to use TART reranker."
			)
		model_name = "facebook/tart-full-flan-t5-xl"
		self.model = EncT5ForSequenceClassification.from_pretrained(model_name)
		self.tokenizer = EncT5Tokenizer.from_pretrained(model_name)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		self.model = self.model.to(self.device)

	def __del__(self):
		del self.model
		del self.tokenizer
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		instruction = kwargs.pop("instruction", "Find passage to answer given question")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, instruction, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		instruction: str = "Find passage to answer given question",
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using Tart.
		TART is a reranker based on TART (https://github.com/facebookresearch/tart).
		You can rerank the passages with the instruction using TARTReranker.
		The default model is facebook/tart-full-flan-t5-xl.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param instruction: The instruction for reranking.
			Note: default instruction is "Find passage to answer given question"
				The default instruction from the TART paper is being used.
				If you want to use a different instruction, you can change the instruction through this parameter
		:param batch: The number of queries to be processed in a batch
		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			[["{} [SEP] {}".format(instruction, query)] for _ in contents]
			for query, contents in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			tart_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
			device=self.device,
			contents_list=contents_list,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def tart_run_model(
	input_texts, contents_list, model, batch_size: int, tokenizer, device
):
	try:
		import torch
		import torch.nn.functional as F
	except ImportError:
		raise ImportError(
			"torch is not installed. Please install torch first to use TART reranker."
		)
	flattened_texts = list(chain.from_iterable(input_texts))
	flattened_contents = list(chain.from_iterable(contents_list))
	batch_input_texts = make_batch(flattened_texts, batch_size)
	batch_contents_list = make_batch(flattened_contents, batch_size)
	results = []
	for batch_texts, batch_contents in zip(batch_input_texts, batch_contents_list):
		feature = tokenizer(
			batch_texts,
			batch_contents,
			padding=True,
			truncation=True,
			return_tensors="pt",
		).to(device)
		with torch.no_grad():
			pred_scores = model(**feature).logits
			normalized_scores = [
				float(score[1]) for score in F.softmax(pred_scores, dim=1)
			]
		results.extend(normalized_scores)
	return results



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Union

import pandas as pd


class BaseModule(metaclass=ABCMeta):
	@abstractmethod
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		pass

	@abstractmethod
	def _pure(self, *args, **kwargs):
		pass

	@classmethod
	def run_evaluator(
		cls,
		project_dir: Union[str, Path],
		previous_result: pd.DataFrame,
		*args,
		**kwargs,
	):
		instance = cls(project_dir, *args, **kwargs)
		result = instance.pure(previous_result, *args, **kwargs)
		del instance
		return result

	@abstractmethod
	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		"""
		This function is for cast function (a.k.a decorator) only for pure function in the whole node.
		"""
		pass



import os
from typing import List, Tuple

import cohere
import pandas as pd
from cohere import RerankResponseResultsItem

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import get_event_loop, process_batch, result_to_dataframe


class CohereReranker(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		"""
		Initialize Cohere rerank node.

		:param project_dir: The project directory path.
		:param api_key: The API key for Cohere rerank.
		    You can set it in the environment variable COHERE_API_KEY.
		    Or, you can directly set it on the config YAML file using this parameter.
		    Default is env variable "COHERE_API_KEY".
		:param kwargs: Extra arguments that are not affected
		"""
		super().__init__(project_dir)
		api_key = kwargs.pop("api_key", None)
		api_key = os.getenv("COHERE_API_KEY", None) if api_key is None else api_key
		if api_key is None:
			api_key = os.getenv("CO_API_KEY", None)
		if api_key is None:
			raise KeyError(
				"Please set the API key for Cohere rerank in the environment variable COHERE_API_KEY "
				"or directly set it on the config YAML file."
			)

		self.cohere_client = cohere.AsyncClientV2(api_key=api_key)

	def __del__(self):
		del self.cohere_client
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		model = kwargs.pop("model", "rerank-v3.5")
		return self._pure(queries, contents, scores, ids, top_k, batch, model)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
		model: str = "rerank-v3.5",
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with Cohere rerank models.
		You can get the API key from https://cohere.com/rerank and set it in the environment variable COHERE_API_KEY.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param scores_list: The list of lists of scores retrieved from the initial ranking
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
		:param model: The model name for Cohere rerank.
		    You can choose between "rerank-v3.5", "rerank-english-v3.0", and "rerank-multilingual-v3.0".
		    Default is "rerank-v3.5".
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		# Run async cohere_rerank_pure function
		tasks = [
			cohere_rerank_pure(self.cohere_client, model, query, document, ids, top_k)
			for query, document, ids in zip(queries, contents_list, ids_list)
		]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
		content_result = list(map(lambda x: x[0], results))
		id_result = list(map(lambda x: x[1], results))
		score_result = list(map(lambda x: x[2], results))

		return content_result, id_result, score_result


async def cohere_rerank_pure(
	cohere_client: cohere.AsyncClient,
	model: str,
	query: str,
	documents: List[str],
	ids: List[str],
	top_k: int,
) -> Tuple[List[str], List[str], List[float]]:
	"""
	Rerank a list of contents with Cohere rerank models.

	:param cohere_client: The Cohere AsyncClient to use for reranking
	:param model: The model name for Cohere rerank
	:param query: The query to use for reranking
	:param documents: The list of contents to rerank
	:param ids: The list of ids corresponding to the documents
	:param top_k: The number of passages to be retrieved
	:return: Tuple of lists containing the reranked contents, ids, and scores
	"""
	rerank_results = await cohere_client.rerank(
		model=model,
		query=query,
		documents=documents,
		top_n=top_k,
	)
	results: List[RerankResponseResultsItem] = rerank_results.results
	reranked_scores: List[float] = list(map(lambda x: x.relevance_score, results))
	indices = list(map(lambda x: x.index, results))
	reranked_contents: List[str] = list(map(lambda i: documents[i], indices))
	reranked_ids: List[str] = list(map(lambda i: ids[i], indices))
	return reranked_contents, reranked_ids, reranked_scores



import json
from pathlib import Path

import pandas as pd
import numpy as np
import os
import zipfile
import requests
from tqdm import tqdm
import collections
from typing import List, Dict, Tuple

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils import result_to_dataframe
from autorag.utils.util import (
	flatten_apply,
	sort_by_scores,
	select_top_k,
	make_batch,
	empty_cuda_cache,
)

model_url = "https://huggingface.co/prithivida/flashrank/resolve/main/{}.zip"

model_file_map = {
	"ms-marco-TinyBERT-L-2-v2": "flashrank-TinyBERT-L-2-v2.onnx",
	"ms-marco-MiniLM-L-12-v2": "flashrank-MiniLM-L-12-v2_Q.onnx",
	"ms-marco-MultiBERT-L-12": "flashrank-MultiBERT-L12_Q.onnx",
	"rank-T5-flan": "flashrank-rankt5_Q.onnx",
	"ce-esci-MiniLM-L12-v2": "flashrank-ce-esci-MiniLM-L12-v2_Q.onnx",
	"miniReranker_arabic_v1": "miniReranker_arabic_v1.onnx",
}


class FlashRankReranker(BasePassageReranker):
	def __init__(
		self, project_dir: str, model: str = "ms-marco-TinyBERT-L-2-v2", *args, **kwargs
	):
		"""
		Initialize FlashRank rerank node.

		:param project_dir: The project directory path.
		:param model: The model name for FlashRank rerank.
		    You can get the list of available models from https://github.com/PrithivirajDamodaran/FlashRank.
		    Default is "ms-marco-TinyBERT-L-2-v2".
		    Not support “rank_zephyr_7b_v1_full” due to parallel inference issue.
		:param kwargs: Extra arguments that are not affected
		"""
		super().__init__(project_dir)
		try:
			from tokenizers import Tokenizer
		except ImportError:
			raise ImportError(
				"Tokenizer is not installed. Please install tokenizers to use FlashRank reranker."
			)

		cache_dir = kwargs.pop("cache_dir", "/tmp")
		max_length = kwargs.pop("max_length", 512)

		self.cache_dir: Path = Path(cache_dir)
		self.model_dir: Path = self.cache_dir / model
		self._prepare_model_dir(model)
		model_file = model_file_map[model]

		try:
			import onnxruntime as ort
		except ImportError:
			raise ImportError(
				"onnxruntime is not installed. Please install onnxruntime to use FlashRank reranker."
			)

		self.session = ort.InferenceSession(str(self.model_dir / model_file))
		self.tokenizer: Tokenizer = self._get_tokenizer(max_length)

	def __del__(self):
		del self.session
		del self.tokenizer
		empty_cuda_cache()
		super().__del__()

	def _prepare_model_dir(self, model_name: str):
		if not self.cache_dir.exists():
			self.cache_dir.mkdir(parents=True, exist_ok=True)

		if not self.model_dir.exists():
			self._download_model_files(model_name)

	def _download_model_files(self, model_name: str):
		local_zip_file = self.cache_dir / f"{model_name}.zip"
		formatted_model_url = model_url.format(model_name)

		with requests.get(formatted_model_url, stream=True) as r:
			r.raise_for_status()
			total_size = int(r.headers.get("content-length", 0))
			with (
				open(local_zip_file, "wb") as f,
				tqdm(
					desc=local_zip_file.name,
					total=total_size,
					unit="iB",
					unit_scale=True,
					unit_divisor=1024,
				) as bar,
			):
				for chunk in r.iter_content(chunk_size=8192):
					size = f.write(chunk)
					bar.update(size)

		with zipfile.ZipFile(local_zip_file, "r") as zip_ref:
			zip_ref.extractall(self.cache_dir)
		os.remove(local_zip_file)

	def _get_tokenizer(self, max_length: int = 512):
		try:
			from tokenizers import AddedToken, Tokenizer
		except ImportError:
			raise ImportError(
				"Pytorch is not installed. Please install pytorch to use FlashRank reranker."
			)
		config = json.load(open(str(self.model_dir / "config.json")))
		tokenizer_config = json.load(
			open(str(self.model_dir / "tokenizer_config.json"))
		)
		tokens_map = json.load(open(str(self.model_dir / "special_tokens_map.json")))
		tokenizer = Tokenizer.from_file(str(self.model_dir / "tokenizer.json"))

		tokenizer.enable_truncation(
			max_length=min(tokenizer_config["model_max_length"], max_length)
		)
		tokenizer.enable_padding(
			pad_id=config["pad_token_id"], pad_token=tokenizer_config["pad_token"]
		)

		for token in tokens_map.values():
			if isinstance(token, str):
				tokenizer.add_special_tokens([token])
			elif isinstance(token, dict):
				tokenizer.add_special_tokens([AddedToken(**token)])

		vocab_file = self.model_dir / "vocab.txt"
		if vocab_file.exists():
			tokenizer.vocab = self._load_vocab(vocab_file)
			tokenizer.ids_to_tokens = collections.OrderedDict(
				[(ids, tok) for tok, ids in tokenizer.vocab.items()]
			)
		return tokenizer

	def _load_vocab(self, vocab_file: Path) -> Dict[str, int]:
		vocab = collections.OrderedDict()
		with open(vocab_file, "r", encoding="utf-8") as reader:
			tokens = reader.readlines()
		for index, token in enumerate(tokens):
			token = token.rstrip("\n")
			vocab[token] = index
		return vocab

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with FlashRank rerank models.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			flashrank_run_model,
			nested_list,
			session=self.session,
			batch_size=batch,
			tokenizer=self.tokenizer,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def flashrank_run_model(input_texts, tokenizer, session, batch_size: int):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []

	for batch_texts in tqdm(batch_input_texts):
		input_text = tokenizer.encode_batch(batch_texts)
		input_ids = np.array([e.ids for e in input_text])
		token_type_ids = np.array([e.type_ids for e in input_text])
		attention_mask = np.array([e.attention_mask for e in input_text])

		use_token_type_ids = token_type_ids is not None and not np.all(
			token_type_ids == 0
		)

		onnx_input = {
			"input_ids": input_ids.astype(np.int64),
			"attention_mask": attention_mask.astype(np.int64),
		}
		if use_token_type_ids:
			onnx_input["token_type_ids"] = token_type_ids.astype(np.int64)

		outputs = session.run(None, onnx_input)

		logits = outputs[0]

		if logits.shape[1] == 1:
			scores = 1 / (1 + np.exp(-logits.flatten()))
		else:
			exp_logits = np.exp(logits)
			scores = exp_logits[:, 1] / np.sum(exp_logits, axis=1)
		results.extend(scores)
	return results



import asyncio
import itertools
import os
from typing import List, Optional

import evaluate
import nltk
import pandas as pd
from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding
from openai import AsyncOpenAI
from pydantic import BaseModel
from rouge_score import tokenizers
from rouge_score.rouge_scorer import RougeScorer
from sacrebleu.metrics.bleu import BLEU

from autorag.embedding.base import embedding_models
from autorag.evaluation.metric.deepeval_prompt import FaithfulnessTemplate
from autorag.evaluation.metric.util import (
	autorag_metric_loop,
	calculate_cosine_similarity,
)
from autorag.nodes.generator import OpenAILLM
from autorag.nodes.generator.base import BaseGenerator
from autorag.schema.metricinput import MetricInput
from autorag.support import get_support_modules
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	openai_truncate_by_token,
	convert_inputs_to_list,
	pop_params,
	empty_cuda_cache,
)


@convert_inputs_to_list
def huggingface_evaluate(
	instance, key: str, metric_inputs: List[MetricInput], **kwargs
) -> List[float]:
	"""
	Compute huggingface evaluate metric.

	:param instance: The instance of huggingface evaluates metric.
	:param key: The key to retrieve result score from huggingface evaluate result.
	:param metric_inputs: A list of MetricInput schema
	:param kwargs: The additional arguments for metric function.
	:return: The list of scores.
	"""

	def compute_score(gt: List[str], pred: str) -> float:
		return max(
			list(
				map(
					lambda x: instance.compute(
						predictions=[pred], references=[x], **kwargs
					)[key],
					gt,
				)
			)
		)

	result = list(
		map(lambda x: compute_score(x.generation_gt, x.generated_texts), metric_inputs)
	)
	return result


def make_generator_instance(generator_module_type: str, llm: str, **kwargs):
	llm_class = get_support_modules(generator_module_type)
	init_params = pop_params(llm_class.__init__, kwargs)
	return llm_class(project_dir="", llm=llm, **init_params)


@autorag_metric_loop(fields_to_check=["retrieval_gt_contents", "generated_texts"])
def deepeval_faithfulness(
	metric_inputs: List[MetricInput],
	generator_module_type: str = "openai_llm",
	lang: str = "en",
	llm: str = "gpt-4o-2024-08-06",
	batch: int = 16,
	**kwargs,
) -> List[float]:
	"""
	Compute deepeval faithfulness metric.
	Its default model is gpt-4o-2024-08-06.
	Since it uses OpenAI model, please be aware of the expensive cost.

	:param metric_inputs: The list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param generator_module_type: Generator module type.
	The default is "openai_llm".
		You can use like "llama_index_llm" or "vllm".
	:param lang: The prompt language that you want to use.
	"en", "ko" and "ja" are supported.
	Korean prompt is not officially supported by DeepEval, but it can be translated by AutoRAG developers.
		Default is "en".
	:param llm: The model name to use for generation.
		Or llm if using llama_index_llm.
		The default is "gpt-4o-2024-08-06".
	:param batch: The batch size for processing.
		Default is 16.
	:param kwargs: The extra parameters for initializing the llm instance.
	:return: The metric scores.
	"""

	class Truth(BaseModel):
		truths: List[str]

	class Claim(BaseModel):
		claims: List[str]

	class Verdict(BaseModel):
		verdict: str
		reason: Optional[str]

	class FaithfulnessVerdicts(BaseModel):
		verdicts: List[Verdict]

	def calculate_score(verdicts: List[Verdict]) -> float:
		number_of_verdicts = len(verdicts)
		if number_of_verdicts == 0:
			return 1

		faithfulness_count = 0
		for verdict in verdicts:
			if verdict.verdict.strip().lower() != "no":
				faithfulness_count += 1

		score = faithfulness_count / number_of_verdicts
		return score

	retrieval_contexts = list(map(lambda x: x.retrieval_gt_contents, metric_inputs))
	truth_prompts = list(
		map(lambda x: FaithfulnessTemplate.generate_truths(x, lang), retrieval_contexts)
	)

	generated_texts = list(map(lambda x: x.generated_texts, metric_inputs))
	claim_prompts = list(
		map(lambda x: FaithfulnessTemplate.generate_claims(x, lang), generated_texts)
	)

	generator: BaseGenerator = make_generator_instance(
		generator_module_type, llm=llm, batch=batch, **kwargs
	)
	if isinstance(generator, OpenAILLM):  # Because of the event loop error at the httpx
		# TODO: Fix the httpx APIConnectionError at the many repetitive request to the OpenAILLM on the same instance
		truth_responses: List[Truth] = generator.structured_output(truth_prompts, Truth)
		claim_responses: List[Claim] = make_generator_instance(
			generator_module_type, llm=llm, batch=batch, **kwargs
		).structured_output(claim_prompts, Claim)
		verdict_prompts = list(
			map(
				lambda claim, truth: FaithfulnessTemplate.generate_verdicts(
					"\n\n".join(claim.claims), "\n\n".join(truth.truths), lang
				),
				claim_responses,
				truth_responses,
			)
		)
		verdict_responses: List[FaithfulnessVerdicts] = make_generator_instance(
			generator_module_type, llm=llm, batch=batch, **kwargs
		).structured_output(verdict_prompts, FaithfulnessVerdicts)
	else:
		truth_responses: List[Truth] = generator.structured_output(truth_prompts, Truth)
		claim_responses: List[Claim] = generator.structured_output(claim_prompts, Claim)
		verdict_prompts = list(
			map(
				lambda claim, truth: FaithfulnessTemplate.generate_verdicts(
					"\n\n".join(claim.claims), "\n\n".join(truth.truths), lang
				),
				claim_responses,
				truth_responses,
			)
		)
		verdict_responses: List[FaithfulnessVerdicts] = generator.structured_output(
			verdict_prompts, FaithfulnessVerdicts
		)

	result = list(map(lambda x: calculate_score(x.verdicts), verdict_responses))
	return result


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def bleu(
	metric_inputs: List[MetricInput],
	tokenize: Optional[str] = None,
	smooth_method: str = "exp",
	smooth_value: Optional[float] = None,
	max_ngram_order: int = 4,
	trg_lang: str = "",
	effective_order: bool = True,
	**kwargs,
) -> List[float]:
	"""
	Computes the BLEU metric given pred and ground-truth.

	:param metric_inputs: A list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param tokenize: The tokenizer to use. If None, defaults to language-specific tokenizers with '13a' as the fallback default. check #https://github.com/mjpost/sacrebleu/blob/master/sacrebleu/metrics/bleu.py
	:param smooth_method: The smoothing method to use ('floor', 'add-k', 'exp' or 'none').
	:param smooth_value: The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
	:param max_ngram_order: If given, it overrides the maximum n-gram order (default: 4) when computing precisions.
	:param trg_lang: An optional language code to raise potential tokenizer warnings.
	:param effective_order: If `True`, stop including n-gram orders for which precision is 0. This should be
	`True`, if sentence-level BLEU will be computed.
	"""
	bleu_instance = BLEU(
		tokenize=tokenize,
		smooth_method=smooth_method,
		smooth_value=smooth_value,
		max_ngram_order=max_ngram_order,
		trg_lang=trg_lang,
		effective_order=effective_order,
		**kwargs,
	)

	result = list(
		map(
			lambda x: bleu_instance.sentence_score(
				x.generated_texts, x.generation_gt
			).score,
			metric_inputs,
		)
	)
	return result


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def meteor(
	metric_inputs: List[MetricInput],
	alpha: float = 0.9,
	beta: float = 3.0,
	gamma: float = 0.5,
) -> List[float]:
	"""
	Compute meteor score for generation.

	:param metric_inputs: A list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param alpha: Parameter for controlling relative weights of precision and recall.
	    Default is 0.9.
	:param beta: Parameter for controlling shape of penalty as a
	    function of as a function of fragmentation.
	    Default is 3.0.
	:param gamma: Relative weight assigned to fragmentation penalty.
	    Default is 0.5.
	:return: A list of computed metric scores.
	"""
	nltk.download("punkt", quiet=True)
	meteor_instance = evaluate.load("meteor")
	result = huggingface_evaluate(
		meteor_instance,
		"meteor",
		metric_inputs,
		alpha=alpha,
		beta=beta,
		gamma=gamma,
	)
	del meteor_instance
	return result


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def rouge(
	metric_inputs: List[MetricInput],
	rouge_type: Optional[str] = "rougeL",
	use_stemmer: bool = False,
	split_summaries: bool = False,
	batch: int = os.cpu_count(),
) -> List[float]:
	"""
	Compute rouge score for generation.

	:param metric_inputs: A list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param rouge_type: A rouge type to use for evaluation.
	    Default is 'RougeL'.
	    Choose between rouge1, rouge2, rougeL, and rougeLSum.
	    - rouge1: unigram (1-gram) based scoring.
	    - rouge2: bigram (2-gram) based scoring.
	    - rougeL: Longest Common Subsequence based scoring.
	    - rougeLSum: splits text using "\n"
	:param use_stemmer: Bool indicating whether Porter stemmer should be used to
	    strip word suffixes to improve matching. This arg is used in the
	    DefaultTokenizer, but other tokenizers might or might not choose to
	    use this. Default is False.
	:param split_summaries: Whether to add newlines between sentences for rougeLsum.
	    Default is False.
	:param batch: The batch size for processing.
	    Default is your cpu count.
	:return: A list of computed metric scores.
	"""
	rouge_instance = RougeScorer(
		rouge_types=[rouge_type],
		use_stemmer=use_stemmer,
		split_summaries=split_summaries,
		tokenizer=tokenizers.DefaultTokenizer(use_stemmer),
	)

	async def compute(gt: List[str], pred: str) -> float:
		return rouge_instance.score_multi(targets=gt, prediction=pred)[
			rouge_type
		].fmeasure

	tasks = [
		compute(metric_input.generation_gt, metric_input.generated_texts)
		for metric_input in metric_inputs
	]
	loop = get_event_loop()
	result = loop.run_until_complete(process_batch(tasks, batch_size=batch))

	del rouge_instance
	return result


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def sem_score(
	metric_inputs: List[MetricInput],
	embedding_model: Optional[BaseEmbedding] = None,
	batch: int = 128,
) -> List[float]:
	"""
	Compute sem score between generation gt and pred with cosine similarity.

	:param metric_inputs: A list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param embedding_model: Embedding model to use for compute cosine similarity.
	    Default is all-mpnet-base-v2 embedding model.
	    The paper used this embedding model.
	:param batch: The batch size for processing.
	    Default is 128
	:return: A list of computed metric scores.
	"""
	generations = [metric_input.generated_texts for metric_input in metric_inputs]
	generation_gt = [metric_input.generation_gt for metric_input in metric_inputs]
	if embedding_model is None:
		embedding_model = embedding_models.get("huggingface_all_mpnet_base_v2")()

	embedding_model.embed_batch_size = batch

	openai_embedding_max_length = 8000
	if isinstance(embedding_model, OpenAIEmbedding):
		generations = openai_truncate_by_token(
			generations, openai_embedding_max_length, embedding_model.model_name
		)

	embedded_pred: List[List[float]] = embedding_model.get_text_embedding_batch(
		generations, show_progress=True
	)
	gt_lengths = list(map(len, generation_gt))
	flatten_gt = list(itertools.chain.from_iterable(generation_gt))
	if isinstance(embedding_model, OpenAIEmbedding):
		flatten_gt = openai_truncate_by_token(
			flatten_gt, openai_embedding_max_length, embedding_model.model_name
		)
	embedded_gt_flatten = embedding_model.get_text_embedding_batch(
		flatten_gt, show_progress=True
	)
	# re-group embedded_gt_flatten with gt_lengths
	iterator = iter(embedded_gt_flatten)
	embedded_gt: List[List[List[float]]] = [
		list(itertools.islice(iterator, length)) for length in gt_lengths
	]

	result = []
	for gt, pred in zip(embedded_gt, embedded_pred):
		similarity_scores: List[float] = list(
			map(lambda x: calculate_cosine_similarity(x, pred), gt)
		)
		result.append(max(similarity_scores))

	del embedding_model
	empty_cuda_cache()

	return result


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def g_eval(
	metric_inputs: List[MetricInput],
	metrics: Optional[List[str]] = None,
	model: str = "gpt-4-0125-preview",
	batch_size: int = 8,
) -> List[float]:
	"""
	Calculate G-Eval score.
	G-eval is a metric that uses high-performance LLM model to evaluate generation performance.
	It evaluates the generation result by coherence, consistency, fluency, and relevance.
	It uses only 'openai' model, and we recommend to use gpt-4 for evaluation accuracy.

	:param metric_inputs: A list of MetricInput schema (Required Field -> "generation_gt", "generated_texts")
	:param metrics: A list of metrics to use for evaluation.
	    Default is all metrics, which is ['coherence', 'consistency', 'fluency', 'relevance'].
	:param model: OpenAI model name.
	    Default is 'gpt-4-0125-preview'.
	:param batch_size: The batch size for processing.
	    Default is 8.
	:return: G-Eval score.
	"""
	generations = [metric_input.generated_texts for metric_input in metric_inputs]
	generation_gt = [metric_input.generation_gt for metric_input in metric_inputs]
	loop = get_event_loop()
	tasks = [
		async_g_eval(gt, pred, metrics, model)
		for gt, pred in zip(generation_gt, generations)
	]
	result = loop.run_until_complete(process_batch(tasks, batch_size=batch_size))
	return result


async def async_g_eval(
	generation_gt: List[str],
	pred: str,
	metrics: Optional[List[str]] = None,
	model: str = "gpt-4-0125-preview",
) -> float:
	available_metrics = ["coherence", "consistency", "fluency", "relevance"]
	if metrics is None:
		metrics = available_metrics
	else:
		assert len(metrics) > 0, "metrics must be a list of string"
		metrics = [metric for metric in metrics if metric in available_metrics]

	current_path = os.path.dirname(os.path.realpath(__file__))
	prompt_path = os.path.join(current_path, "g_eval_prompts")
	g_eval_prompts = {
		"coherence": open(os.path.join(prompt_path, "coh_detailed.txt")).read(),
		"consistency": open(os.path.join(prompt_path, "con_detailed.txt")).read(),
		"fluency": open(os.path.join(prompt_path, "flu_detailed.txt")).read(),
		"relevance": open(os.path.join(prompt_path, "rel_detailed.txt")).read(),
	}

	client = AsyncOpenAI()

	async def g_eval_score(prompt: str, gen_gt: List[str], pred: str):
		scores = []
		for gt in gen_gt:
			input_prompt = prompt.replace("{{Document}}", gt).replace(
				"{{Summary}}", pred
			)
			response = await client.chat.completions.create(
				model=model,
				messages=[{"role": "system", "content": input_prompt}],
				logprobs=True,
				top_logprobs=5,
				temperature=0,
				max_tokens=2,
				frequency_penalty=0,
				presence_penalty=0,
				stop=None,
				n=20,
			)
			if "(1-3):" in prompt:
				scores.append(get_g_eval_score(response, max_score=3))
			else:
				scores.append(get_g_eval_score(response))
		return max(scores)

	def get_g_eval_score(responses, max_score: int = 5) -> int:
		target_tokens = {str(i): 0 for i in range(1, max_score + 1)}
		for choice in responses.choices:
			first_top_log_probs = choice.logprobs.content[0].top_logprobs
			for i, top_log_prob in enumerate(
				list(map(lambda x: x.token, first_top_log_probs))
			):
				if top_log_prob in target_tokens:
					target_tokens[top_log_prob] += 5 - i

		return int(max(target_tokens, key=target_tokens.get))

	g_eval_scores = await asyncio.gather(
		*(g_eval_score(g_eval_prompts[x], generation_gt, pred) for x in metrics)
	)
	return sum(g_eval_scores) / len(g_eval_scores)


@autorag_metric_loop(fields_to_check=["generation_gt", "generated_texts"])
def bert_score(
	metric_inputs: List[MetricInput],
	lang: str = "en",
	batch: int = 128,
	n_threads: int = os.cpu_count(),
) -> List[float]:
	generations = [metric_input.generated_texts for metric_input in metric_inputs]
	generation_gt = [metric_input.generation_gt for metric_input in metric_inputs]
	evaluator = evaluate.load("bertscore")

	df = pd.DataFrame(
		{
			"reference": generation_gt,
			"prediction": generations,
			"lang": lang,
		}
	)

	df = df.explode("reference", ignore_index=False)
	df["bert_score"] = evaluator.compute(
		predictions=df["prediction"].tolist(),
		references=df["reference"].tolist(),
		lang=lang,
		nthreads=n_threads,
		batch_size=batch,
	)["f1"]

	del evaluator
	empty_cuda_cache()

	return df.groupby(level=0)["bert_score"].max().tolist()



# The RAGAS prompts are coming from RAGAS under Apache-2.0 License. (English version) (the AutoRAG team translates Korean version prompt)
# You can see the original prompts at the RAGAS library at https://github.com/explodinggradients/ragas/blob/main/src/ragas/testset/prompts.py
from llama_index.core.base.llms.types import ChatMessage, MessageRole

QUERY_EVOLVE_PROMPT = {
	"conditional_evolve_ragas": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""Rewrite the provided question to increase its complexity by introducing a conditional element.
The goal is to make the question more intricate by incorporating a scenario or condition that affects the context of the question.
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present context.
    4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.
""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question : What is the function of the roots of a plant?
Context : The roots of a plant absorb water and nutrients from the soil, anchor the plant in the ground, and store food.
Output : """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="What dual purpose do plant roots serve concerning soil nutrients and stability?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question : How do vaccines protect against diseases?
Context : Vaccines protect against diseases by stimulating the body's immune response to produce antibodies, which recognize and combat pathogens.
Output : """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="How do vaccines utilize the body's immune system to defend against pathogens?",
			),
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""제공된 질문에 조건에 관련한 내용을 추가하여 복잡성을 높이세요.
질문의 Context에 영향을 미치는 시나리오나 조건을 포함하여 질문을 더 복잡하게 만드는 것이 목표입니다.
질문을 다시 작성할 때 다음 규칙을 따르십시오.
    1. 다시 작성된 질문은 100자를 넘지 않아야 합니다. 가능한 경우 약어를 사용하십시오.
    2. 다시 작성된 질문은 합리적이어야 하며 사람이 이해하고 응답할 수 있어야 합니다.
    3. 다시 작성된 질문은 현재 Context에서 완전히 답변할 수 있어야 합니다.
    4. '제공된 글', '단락에 따르면?', 'Context에 의하면' 등의 문구는 질문에 나타날 수 없습니다.
    5. 한국어로 질문을 작성하세요.
""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 식물의 뿌리 기능이 뭐야?
Context: 식물의 뿌리는 토양에서 물과 영양분을 흡수하고, 식물을 땅에 고정하며, 영양분을 저장합니다.
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="식물의 뿌리는 토양 영양분과 안정성에 대해 어떤 역할을 하나요?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 백신은 질병을 어떻게 예방하나요?
Context: 백신은 신체의 면역 반응을 자극하여 병원체를 인식하고 싸우는 항체를 생성함으로써 질병으로부터 보호합니다.
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="백신은 신체의 면역 체계를 어떻게 활용해서 질병을 예방합니까?",
			),
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""提供された質問に条件に関する内容を追加して、複雑さを高めます。
質問のContextに影響を与えるシナリオや条件を含めて、質問をより複雑にすることが目標です。
質問を再作成するときは、次のルールに従います。
    1. 再作成された質問は100文字を超えてはいけません。 可能であれば略語を使ってください
    2. 再作成された質問は合理的でなければならず、人が理解して回答できるものでなければなりません。
    3. 再作成された質問は、現在のContextで完全に答えられる必要があります。
    4. 「提供された文」、「段落によると?」、「Contextによると」などのフレーズは質問に表示されません。
    5. 日本語で質問を書きましょう。
""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 植物の根の機能は何ですか？
Context: 植物の根は土壌から水や栄養分を吸収し、植物を地面に固定し、栄養分を蓄えます。
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="植物の根は土壌栄養分と安定性に対してどのような役割をしますか？",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: ワクチンは病気をどのように予防しますか?
Context: ワクチンは、体の免疫反応を刺激して病原体を認識し、戦う抗体を生成することで病気から守ります。
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="ワクチンは体の免疫システムをどのように活用して病気を予防しますか？",
			),
		],
	},
	"reasoning_evolve_ragas": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""Complicate the given question by rewriting question into a multi-hop reasoning question based on the provided context.
Answering the question should require the reader to make multiple logical connections or inferences using the information available in given context.
Rules to follow when rewriting question:
1. Ensure that the rewritten question can be answered entirely from the information present in the contexts.
2. Do not frame questions that contains more than 15 words. Use abbreviation wherever possible.
3. Make sure the question is clear and unambiguous.
4. phrases like 'based on the provided context','according to the context',etc are not allowed to appear in the question.""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: What is the capital of France?,
Context: France is a country in Western Europe. It has several cities, including Paris, Lyon, and Marseille. Paris is not only known for its cultural landmarks like the Eiffel Tower and the Louvre Museum but also as the administrative center.
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="Linking the Eiffel Tower and administrative center, which city stands as both?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: What does the append() method do in Python?
Context: In Python, lists are used to store multiple items in a single variable. Lists are one of 4 built-in data types used to store collections of data. The append() method adds a single item to the end of a list.
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="If a list represents a variable collection, what method extends it by one item?",
			),
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""주어진 Context를 기반으로 기존 질문을 복잡하게 만들어 여러 논리적인 사고가 필요한 질문으로 다시 작성하세요.
질문에 답하려면 주어진 Context의 정보를 사용해 여러 논리적 사고나 추론을 해야 합니다.
질문을 다시 작성할 때 따라야 할 규칙:
1. 다시 작성된 질문은 Context에 있는 정보만으로 완전히 답변할 수 있어야 합니다.
2. 100자를 초과하는 질문을 작성하지 마세요. 가능한 경우 약어를 사용하세요.
3. 질문이 명확하고 모호하지 않도록 하세요.
4. '제공된 Context에 기반하여', '해당 단락에 따르면' 등의 문구는 질문에 포함되지 않아야 합니다.
5. 한국어로 질문을 작성하세요.""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 프랑스의 수도는 어디인가요?,
Context: 프랑스는 서유럽에 있는 나라입니다. 파리, 리옹, 마르세유를 포함한 여러 도시가 있습니다. 파리는 에펠탑과 루브르 박물관 같은 문화적 랜드마크로 유명할 뿐만 아니라 행정 중심지로도 알려져 있습니다.
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="에펠탑과 행정 중심지, 두 단어는 어떤 도시를 가리키나요?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""질문: Python에서 append() 메서드는 무엇을 하나요?
컨텍스트: Python에서 리스트는 하나의 변수에 여러 항목을 저장하는 데 사용됩니다. 리스트는 데이터를 저장하는 데 사용되는 4가지 내장 데이터 유형 중 하나입니다. append() 메서드는 리스트의 끝에 새로운 항목을 추가합니다.
출력: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="리스트가 변수들을 모아 놓은 것을 나타낸다면, 어떤 메서드를 사용해야 항목을 하나 더 추가할 수 있습니까?",
			),
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""与えられたContextに基づいて既存の質問を複雑にして、様々な論理的思考が必要な質問として書き直しましょう。
質問に答えるためには、与えられたContextの情報を使って様々な論理的思考や推論をしなければなりません。
質問を再作成するときに従うべきルール:
1. 再作成された質問は、Contextにある情報だけで完全に答えられる必要があります。
2. 100文字を超える質問を作成してはいけません。 可能であれば略語を使ってください。
3. 質問が明確で曖昧にならないようにしましょう。
4. 「提供されたContextに基づいて」、「当該段落によると」などのフレーズは、質問に含まれてはいけません。
5. 日本語で質問を書きましょう。""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: フランスの首都はどこですか？,
Context: フランスは西ヨーロッパにある国です。 パリ、リヨン、マルセイユを含むいくつかの都市があります。 パリはエッフェル塔やルーブル博物館のような文化的ランドマークとして有名なだけでなく、行政の中心地としても知られています。
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="エッフェル塔と行政の中心地、二つの単語はどんな都市を指していますか？",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: Pythonでappend() メソッドは何をしますか？
Context: Pythonで、リストは 1 つの変数に複数の項目を保存するために使用されます。 リストは、データを保存するために使用される 4 つの組み込みデータ タイプの 1 つです。 append()メソッドは、リストの最後に新しい項目を追加します。
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="リストが変数を集めたものである場合、どのメソッドを使えば項目を一つ追加することができますか？",
			),
		],
	},
	"compress_ragas": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""Rewrite the following question to make it more indirect and shorter while retaining the essence of the original question.
    The goal is to create a question that conveys the same meaning but in a less direct manner. The rewritten question should shorter so use abbreviation wherever possible.""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: What is the distance between the Earth and the Moon?
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="How far is the Moon from Earth?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: What ingredients are required to bake a chocolate cake?
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="What's needed for a chocolate cake?",
			),
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""주어진 질문을 더 간접적이고 짧게 다시 작성하세요.
        목표는 질문을 원래 질문의 본질을 유지하면서 너무 직설적이지 않게 만드는 것입니다.
        약어 등을 사용하여 질문을 더 짧게 만드세요.""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 지구와 달 사이의 거리는 얼마입니까?
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="달은 지구에서 얼마나 떨어져 있나요?",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 초콜릿 케이크를 굽기 위해 필요한 재료는 무엇입니까?
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="초콜릿 케이크에 필요한 것은 무엇인가요?",
			),
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""与えられた質問をより間接的かつ短く書き換えます。
目標は、質問を元の質問の本質を保ちながら、あまりストレートにならないようにすることです。
略語などを使用して、質問をより短くします。""",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: 地球と月の間の距離はどれくらいですか？
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="月は地球からどれくらい離れていますか？",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Question: チョコレートケーキを焼くために必要な材料は何ですか？
Output: """,
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="チョコレートケーキに必要なものは何ですか？",
			),
		],
	},
}



import abc
import logging
from pathlib import Path
from typing import List, Union

import pandas as pd

from autorag.nodes.util import make_generator_callable_param
from autorag.schema import BaseModule
from autorag.utils import validate_qa_dataset

logger = logging.getLogger("AutoRAG")


class BaseQueryExpansion(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		logger.info(
			f"Initialize query expansion node - {self.__class__.__name__} module..."
		)
		# set generator module for query expansion
		generator_class, generator_param = make_generator_callable_param(kwargs)
		self.generator = generator_class(project_dir, **generator_param)

	def __del__(self):
		del self.generator
		logger.info(
			f"Delete query expansion node - {self.__class__.__name__} module..."
		)

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(
			f"Running query expansion node - {self.__class__.__name__} module..."
		)
		validate_qa_dataset(previous_result)

		# find queries columns
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		queries = previous_result["query"].tolist()
		return queries

	@staticmethod
	def _check_expanded_query(queries: List[str], expanded_queries: List[List[str]]):
		return list(
			map(
				lambda query, expanded_query_list: check_expanded_query(
					query, expanded_query_list
				),
				queries,
				expanded_queries,
			)
		)


def check_expanded_query(query: str, expanded_query_list: List[str]):
	# check if the expanded query is the same as the original query
	expanded_query_list = list(map(lambda x: x.strip(), expanded_query_list))
	return [
		expanded_query if expanded_query else query
		for expanded_query in expanded_query_list
	]



import logging
import uuid
from typing import Callable, Optional, List

import chromadb
import numpy as np
import pandas as pd
from tqdm import tqdm

import autorag
from autorag.nodes.semanticretrieval.vectordb import vectordb_ingest_api, vectordb_pure
from autorag.utils.util import (
	save_parquet_safe,
	fetch_contents,
	get_event_loop,
	process_batch,
)

logger = logging.getLogger("AutoRAG")


def make_single_content_qa(
	corpus_df: pd.DataFrame,
	content_size: int,
	qa_creation_func: Callable,
	output_filepath: Optional[str] = None,
	upsert: bool = False,
	random_state: int = 42,
	cache_batch: int = 32,
	**kwargs,
) -> pd.DataFrame:
	"""
	Make single content (single-hop, single-document) QA dataset using given qa_creation_func.
	It generates a single content QA dataset, which means its retrieval ground truth will be only one.
	It is the most basic form of QA dataset.

	:param corpus_df: The corpus dataframe to make QA dataset from.
	:param content_size: This function will generate QA dataset for the given number of contents.
	:param qa_creation_func: The function to create QA pairs.
	    You can use like `generate_qa_llama_index` or `generate_qa_llama_index_by_ratio`.
	    The input func must have `contents` parameter for the list of content string.
	:param output_filepath: Optional filepath to save the parquet file.
	    If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
	    File directory must exist. File extension must be .parquet
	:param upsert: If true, the function will overwrite the existing file if it exists.
	    Default is False.
	:param random_state: The random state for sampling corpus from the given corpus_df.
	:param cache_batch: The number of batches to use for caching the generated QA dataset.
	    When the cache_batch size data is generated, the dataset will save to the designated output_filepath.
	    If the cache_batch size is too small, the process time will be longer.
	:param kwargs: The keyword arguments for qa_creation_func.
	:return: QA dataset dataframe.
	    You can save this as parquet file to use at AutoRAG.
	"""
	assert content_size > 0, "content_size must be greater than 0."
	if content_size > len(corpus_df):
		logger.warning(
			f"content_size {content_size} is larger than the corpus size {len(corpus_df)}. "
			"Setting content_size to the corpus size."
		)
		content_size = len(corpus_df)
	sampled_corpus = corpus_df.sample(n=content_size, random_state=random_state)
	sampled_corpus = sampled_corpus.reset_index(drop=True)

	def make_query_generation_gt(row):
		return row["qa"]["query"], row["qa"]["generation_gt"]

	qa_data = pd.DataFrame()
	for idx, i in tqdm(enumerate(range(0, len(sampled_corpus), cache_batch))):
		qa = qa_creation_func(
			contents=sampled_corpus["contents"].tolist()[i : i + cache_batch], **kwargs
		)

		temp_qa_data = pd.DataFrame(
			{
				"qa": qa,
				"retrieval_gt": sampled_corpus["doc_id"].tolist()[i : i + cache_batch],
			}
		)
		temp_qa_data = temp_qa_data.explode("qa", ignore_index=True)
		temp_qa_data["qid"] = [str(uuid.uuid4()) for _ in range(len(temp_qa_data))]
		temp_qa_data[["query", "generation_gt"]] = temp_qa_data.apply(
			make_query_generation_gt, axis=1, result_type="expand"
		)
		temp_qa_data = temp_qa_data.drop(columns=["qa"])

		temp_qa_data["retrieval_gt"] = temp_qa_data["retrieval_gt"].apply(
			lambda x: [[x]]
		)
		temp_qa_data["generation_gt"] = temp_qa_data["generation_gt"].apply(
			lambda x: [x]
		)

		if idx == 0:
			qa_data = temp_qa_data
		else:
			qa_data = pd.concat([qa_data, temp_qa_data], ignore_index=True)
		if output_filepath is not None:
			save_parquet_safe(qa_data, output_filepath, upsert=upsert)

	return qa_data


def make_qa_with_existing_qa(
	corpus_df: pd.DataFrame,
	existing_query_df: pd.DataFrame,
	content_size: int,
	answer_creation_func: Optional[Callable] = None,
	exist_gen_gt: Optional[bool] = False,
	output_filepath: Optional[str] = None,
	embedding_model: str = "openai_embed_3_large",
	collection: Optional[chromadb.Collection] = None,
	upsert: bool = False,
	random_state: int = 42,
	cache_batch: int = 32,
	top_k: int = 3,
	**kwargs,
) -> pd.DataFrame:
	"""
	Make single-hop QA dataset using given qa_creation_func and existing queries.

	:param corpus_df: The corpus dataframe to make QA dataset from.
	:param existing_query_df: Dataframe containing existing queries to use for QA pair creation.
	:param content_size: This function will generate QA dataset for the given number of contents.
	:param answer_creation_func: Optional function to create answer with input query.
	    If exist_gen_gt is False, this function must be given.
	:param exist_gen_gt: Optional boolean to use existing generation_gt.
	    If True, the existing_query_df must have 'generation_gt' column.
	    If False, the answer_creation_func must be given.
	:param output_filepath: Optional filepath to save the parquet file.
	:param embedding_model: The embedding model to use for vectorization.
	    You can add your own embedding model in the autorag.embedding_models.
	    Please refer to how to add an embedding model in this doc: https://marker-inc-korea.github.io/AutoRAG/local_model.html
	    The default is 'openai_embed_3_large'.
	:param collection: The chromadb collection to use for vector DB.
	    You can make any chromadb collection and use it here.
	    If you already ingested the corpus_df to the collection, the embedding process will not be repeated.
	    The default is None. If None, it makes a temporary collection.
	:param upsert: If true, the function will overwrite the existing file if it exists.
	:param random_state: The random state for sampling corpus from the given corpus_df.
	:param cache_batch: The number of batches to use for caching the generated QA dataset.
	:param top_k: The number of sources to refer by model.
	    Default is 3.
	:param kwargs: The keyword arguments for qa_creation_func.
	:return: QA dataset dataframe.
	"""
	raise DeprecationWarning("This function is deprecated.")
	assert (
		"query" in existing_query_df.columns
	), "existing_query_df must have 'query' column."

	if exist_gen_gt:
		assert (
			"generation_gt" in existing_query_df.columns
		), "existing_query_df must have 'generation_gt' column."
	else:
		assert (
			answer_creation_func is not None
		), "answer_creation_func must be given when exist_gen_gt is False."

	assert content_size > 0, "content_size must be greater than 0."
	if content_size > len(corpus_df):
		logger.warning(
			f"content_size {content_size} is larger than the corpus size {len(corpus_df)}. "
			"Setting content_size to the corpus size."
		)
		content_size = len(corpus_df)

	logger.info("Loading local embedding model...")
	embeddings = autorag.embedding_models[embedding_model]()

	# Vector DB creation
	if collection is None:
		chroma_client = chromadb.Client()
		collection_name = "auto-rag"
		collection = chroma_client.get_or_create_collection(collection_name)

	# embed corpus_df
	vectordb_ingest_api(collection, corpus_df, embeddings)
	query_embeddings = embeddings.get_text_embedding_batch(
		existing_query_df["query"].tolist()
	)

	loop = get_event_loop()
	tasks = [
		vectordb_pure([query_embedding], top_k, collection)
		for query_embedding in query_embeddings
	]
	results = loop.run_until_complete(process_batch(tasks, batch_size=cache_batch))
	retrieved_ids = list(map(lambda x: x[0], results))

	retrieved_contents: List[List[str]] = fetch_contents(corpus_df, retrieved_ids)
	input_passage_strs: List[str] = list(
		map(
			lambda x: "\n".join(
				[f"Document {i + 1}\n{content}" for i, content in enumerate(x)]
			),
			retrieved_contents,
		)
	)

	retrieved_qa_df = pd.DataFrame(
		{
			"qid": [str(uuid.uuid4()) for _ in range(len(existing_query_df))],
			"query": existing_query_df["query"].tolist(),
			"retrieval_gt": list(map(lambda x: [x], retrieved_ids)),
			"input_passage_str": input_passage_strs,
		}
	)

	if exist_gen_gt:
		generation_gt = existing_query_df["generation_gt"].tolist()
		if isinstance(generation_gt[0], np.ndarray):
			retrieved_qa_df["generation_gt"] = generation_gt
		else:
			raise ValueError(
				"In existing_query_df, generation_gt (per query) must be in the form of List[str]."
			)

	sample_qa_df = retrieved_qa_df.sample(
		n=min(content_size, len(retrieved_qa_df)), random_state=random_state
	)

	qa_df = sample_qa_df.copy(deep=True)
	qa_df.drop(columns=["input_passage_str"], inplace=True)

	if not exist_gen_gt:
		generation_gt = answer_creation_func(
			contents=sample_qa_df["input_passage_str"].tolist(),
			queries=sample_qa_df["query"].tolist(),
			batch=cache_batch,
			**kwargs,
		)
		qa_df["generation_gt"] = generation_gt

	if output_filepath is not None:
		save_parquet_safe(qa_df, output_filepath, upsert=upsert)

	return qa_df



import logging
import logging.config
import os
import sys
from random import random
from typing import List, Any

from llama_index.core.embeddings.mock_embed_model import MockEmbedding
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.core.llms.mock import MockLLM
from llama_index.llms.bedrock import Bedrock
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.embeddings.openai import OpenAIEmbeddingModelType

from llama_index.llms.openai import OpenAI
from llama_index.llms.openai_like import OpenAILike
from langchain_openai.embeddings import OpenAIEmbeddings
from rich.logging import RichHandler

version_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "VERSION")

with open(version_path, "r") as f:
	__version__ = f.read().strip()


class LazyInit:
	def __init__(self, factory, *args, **kwargs):
		self._factory = factory
		self._args = args
		self._kwargs = kwargs
		self._instance = None

	def __call__(self):
		if self._instance is None:
			self._instance = self._factory(*self._args, **self._kwargs)
		return self._instance

	def __getattr__(self, name):
		if self._instance is None:
			self._instance = self._factory(*self._args, **self._kwargs)
		return getattr(self._instance, name)


rich_format = "[%(filename)s:%(lineno)s] >> %(message)s"
logging.basicConfig(
	level="INFO", format=rich_format, handlers=[RichHandler(rich_tracebacks=True)]
)
logger = logging.getLogger("AutoRAG")


def handle_exception(exc_type, exc_value, exc_traceback):
	logger = logging.getLogger("AutoRAG")
	logger.error("Unexpected exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception


class AutoRAGBedrock(Bedrock):
	async def acomplete(
		self, prompt: str, formatted: bool = False, **kwargs: Any
	) -> CompletionResponse:
		return self.complete(prompt, formatted=formatted, **kwargs)


generator_models = {
	"openai": OpenAI,
	"openailike": OpenAILike,
	"mock": MockLLM,
	"bedrock": AutoRAGBedrock,
}

try:
	from llama_index.llms.huggingface import HuggingFaceLLM
	from llama_index.llms.ollama import Ollama

	generator_models["huggingfacellm"] = HuggingFaceLLM
	generator_models["ollama"] = Ollama

except ImportError:
	logger.info(
		"You are using API version of AutoRAG."
		"To use local version, run pip install 'AutoRAG[gpu]'"
	)

try:
	import transformers

	transformers.logging.set_verbosity_error()
except ImportError:
	logger.info(
		"You are using API version of AutoRAG."
		"To use local version, run pip install 'AutoRAG[gpu]'"
	)



import os
import tempfile
from glob import glob
from typing import List, Tuple, Dict

from PyPDF2 import PdfFileReader, PdfFileWriter
import pdfplumber

from autorag.support import get_support_modules
from autorag.data.parse.base import parser_node


@parser_node
def table_hybrid_parse(
	data_path_list: List[str],
	text_parse_module: str,
	text_params: Dict,
	table_parse_module: str,
	table_params: Dict,
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parse documents to use table_hybrid_parse method.
	The table_hybrid_parse method is a hybrid method that combines the parsing results of PDFs with and without tables.
	It splits the PDF file into pages, separates pages with and without tables, and then parses and merges the results.

	:param data_path_list: The list of data paths to parse.
	:param text_parse_module: The text parsing module to use. The type should be a string.
	:param text_params: The extra parameters for the text parsing module. The type should be a dictionary.
	:param table_parse_module: The table parsing module to use. The type should be a string.
	:param table_params: The extra parameters for the table parsing module. The type should be a dictionary.
	:return: tuple of lists containing the parsed texts, path and pages.
	"""
	# make save folder directory
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as save_dir:
		text_dir = os.path.join(save_dir, "text")
		table_dir = os.path.join(save_dir, "table")

		os.makedirs(text_dir, exist_ok=True)
		os.makedirs(table_dir, exist_ok=True)

		# Split PDF file into pages and Save PDFs with and without tables
		path_map_dict_lst = [
			save_page_by_table(data_path, text_dir, table_dir)
			for data_path in data_path_list
		]
		path_map_dict = {k: v for d in path_map_dict_lst for k, v in d.items()}

		# Extract text pages
		table_results, table_file_path = get_each_module_result(
			table_parse_module, table_params, os.path.join(table_dir, "*")
		)

		# Extract table pages
		text_results, text_file_path = get_each_module_result(
			text_parse_module, text_params, os.path.join(text_dir, "*")
		)

		# Merge parsing results of PDFs with and without tables
		texts = table_results + text_results
		temp_path_lst = table_file_path + text_file_path

		# Sort by file names
		temp_path_lst, texts = zip(*sorted(zip(temp_path_lst, texts)))

		# get original file path
		path = list(map(lambda temp_path: path_map_dict[temp_path], temp_path_lst))

		# get pages
		pages = list(map(lambda x: get_page_from_path(x), temp_path_lst))

		return list(texts), path, pages


# Save PDFs with and without tables
def save_page_by_table(data_path: str, text_dir: str, table_dir: str) -> Dict[str, str]:
	file_name = os.path.basename(data_path).split(".pdf")[0]

	with open(data_path, "rb") as input_data:
		pdf_reader = PdfFileReader(input_data)
		num_pages = pdf_reader.getNumPages()

		path_map_dict = {}
		for page_num in range(num_pages):
			output_pdf_path = _get_output_path(
				data_path, page_num, file_name, text_dir, table_dir
			)
			_save_single_page(pdf_reader, page_num, output_pdf_path)
			path_map_dict.update({output_pdf_path: data_path})

	return path_map_dict


def _get_output_path(
	data_path: str, page_num: int, file_name: str, text_dir: str, table_dir: str
) -> str:
	with pdfplumber.open(data_path) as pdf:
		page = pdf.pages[page_num]
		tables = page.extract_tables()
		directory = table_dir if tables else text_dir
		return os.path.join(directory, f"{file_name}_page_{page_num + 1}.pdf")


def _save_single_page(pdf_reader: PdfFileReader, page_num: int, output_pdf_path: str):
	pdf_writer = PdfFileWriter()
	pdf_writer.addPage(pdf_reader.getPage(page_num))

	with open(output_pdf_path, "wb") as output_file:
		pdf_writer.write(output_file)


def get_each_module_result(
	module: str, module_params: Dict, data_path_glob: str
) -> Tuple[List[str], List[str]]:
	module_params["module_type"] = module

	data_path_list = glob(data_path_glob)
	if not data_path_list:
		return [], []

	module_name = module_params.pop("module_type")
	module_callable = get_support_modules(module_name)
	module_original = module_callable.__wrapped__
	texts, path, _ = module_original(data_path_list, **module_params)

	return texts, path


def get_page_from_path(file_path: str) -> int:
	file_name = os.path.basename(file_path)
	split_result = file_name.rsplit("_page_", -1)
	page_number_with_extension = split_result[1]
	page_number, _ = page_number_with_extension.split(".")

	return int(page_number)



import functools
from typing import List

import numpy as np

from autorag.schema.metricinput import MetricInput
from autorag.utils.util import convert_inputs_to_list


def calculate_cosine_similarity(a, b):
	dot_product = np.dot(a, b)
	norm_a = np.linalg.norm(a)
	norm_b = np.linalg.norm(b)
	similarity = dot_product / (norm_a * norm_b)
	return similarity


def calculate_l2_distance(a, b):
	return np.linalg.norm(a - b)


def calculate_inner_product(a, b):
	return np.dot(a, b)


def autorag_metric(fields_to_check: List[str]):
	def decorator_autorag_metric(func):
		@functools.wraps(func)
		@convert_inputs_to_list
		def wrapper(metric_inputs: List[MetricInput], **kwargs) -> List[float]:
			"""
			Use 'for loop' to run each metric input.
			Put the single metric input into the metric function.

			:param metric_inputs: A list MetricInput schema for AutoRAG metric.
			:param kwargs: The additional arguments for metric function.
			:return: A list of computed metric scores.
			"""
			results = []
			for metric_input in metric_inputs:
				if metric_input.is_fields_notnone(fields_to_check=fields_to_check):
					results.append(func(metric_input, **kwargs))
				else:
					results.append(None)
			return results

		return wrapper

	return decorator_autorag_metric


def autorag_metric_loop(fields_to_check: List[str]):
	def decorator_autorag_generation_metric(func):
		@functools.wraps(func)
		@convert_inputs_to_list
		def wrapper(metric_inputs: List[MetricInput], **kwargs) -> List[float]:
			"""
			Put the list of metric inputs into the metric function.

			:param metric_inputs: A list MetricInput schema for AutoRAG metric.
			:param kwargs: The additional arguments for metric function.
			:return: A list of computed metric scores.
			"""
			bool_list = [
				metric_input.is_fields_notnone(fields_to_check=fields_to_check)
				for metric_input in metric_inputs
			]
			valid_inputs = [
				metric_input
				for metric_input, is_valid in zip(metric_inputs, bool_list)
				if is_valid
			]

			results = [None] * len(metric_inputs)
			if valid_inputs:
				processed_valid = func(valid_inputs, **kwargs)

				valid_index = 0
				for i, is_valid in enumerate(bool_list):
					if is_valid:
						results[i] = processed_valid[valid_index]
						valid_index += 1

			return results

		return wrapper

	return decorator_autorag_generation_metric



import itertools
import math

from autorag.evaluation.metric.util import autorag_metric
from autorag.schema.metricinput import MetricInput


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_f1(metric_input: MetricInput):
	"""
	Compute f1 score for retrieval.

	:param metric_input: The MetricInput schema for AutoRAG metric.
	:return: The f1 score.
	"""
	recall_score = retrieval_recall.__wrapped__(metric_input)
	precision_score = retrieval_precision.__wrapped__(metric_input)
	if recall_score + precision_score == 0:
		return 0
	else:
		return 2 * (recall_score * precision_score) / (recall_score + precision_score)


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_recall(metric_input: MetricInput) -> float:
	gt, pred = metric_input.retrieval_gt, metric_input.retrieved_ids

	gt_sets = [frozenset(g) for g in gt]
	pred_set = set(pred)
	hits = sum(any(pred_id in gt_set for pred_id in pred_set) for gt_set in gt_sets)
	recall = hits / len(gt) if len(gt) > 0 else 0.0
	return recall


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_precision(metric_input: MetricInput) -> float:
	gt, pred = metric_input.retrieval_gt, metric_input.retrieved_ids

	gt_sets = [frozenset(g) for g in gt]
	pred_set = set(pred)
	hits = sum(any(pred_id in gt_set for gt_set in gt_sets) for pred_id in pred_set)
	precision = hits / len(pred) if len(pred) > 0 else 0.0
	return precision


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_ndcg(metric_input: MetricInput) -> float:
	gt, pred = metric_input.retrieval_gt, metric_input.retrieved_ids

	gt_sets = [frozenset(g) for g in gt]
	pred_set = set(pred)
	relevance_scores = {
		pred_id: 1 if any(pred_id in gt_set for gt_set in gt_sets) else 0
		for pred_id in pred_set
	}

	dcg = sum(
		(2 ** relevance_scores[doc_id] - 1) / math.log2(i + 2)
		for i, doc_id in enumerate(pred)
	)

	len_flatten_gt = len(list(itertools.chain.from_iterable(gt)))
	len_pred = len(pred)
	ideal_pred = [1] * min(len_flatten_gt, len_pred) + [0] * max(
		0, len_pred - len_flatten_gt
	)
	idcg = sum(relevance / math.log2(i + 2) for i, relevance in enumerate(ideal_pred))

	ndcg = dcg / idcg if idcg > 0 else 0
	return ndcg


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_mrr(metric_input: MetricInput) -> float:
	"""
	Reciprocal Rank (RR) is the reciprocal of the rank of the first relevant item.
	Mean of RR in whole queries is MRR.
	"""
	gt, pred = metric_input.retrieval_gt, metric_input.retrieved_ids

	# Flatten the ground truth list of lists into a single set of relevant documents
	gt_sets = [frozenset(g) for g in gt]

	rr_list = []
	for gt_set in gt_sets:
		for i, pred_id in enumerate(pred):
			if pred_id in gt_set:
				rr_list.append(1.0 / (i + 1))
				break
	return sum(rr_list) / len(gt_sets) if rr_list else 0.0


@autorag_metric(fields_to_check=["retrieval_gt", "retrieved_ids"])
def retrieval_map(metric_input: MetricInput) -> float:
	"""
	Mean Average Precision (MAP) is the mean of Average Precision (AP) for all queries.
	"""
	gt, pred = metric_input.retrieval_gt, metric_input.retrieved_ids

	gt_sets = [frozenset(g) for g in gt]

	ap_list = []

	for gt_set in gt_sets:
		pred_hits = [1 if pred_id in gt_set else 0 for pred_id in pred]
		precision_list = [
			sum(pred_hits[: i + 1]) / (i + 1)
			for i, hit in enumerate(pred_hits)
			if hit == 1
		]
		ap_list.append(
			sum(precision_list) / len(precision_list) if precision_list else 0.0
		)

	return sum(ap_list) / len(gt_sets) if ap_list else 0.0



from typing import List, Optional

import pandas as pd

from autorag.nodes.passagecompressor.base import BasePassageCompressor
from autorag.utils.util import pop_params, result_to_dataframe, empty_cuda_cache


# TODO: Parallel Processing Refactoring at #460


class LongLLMLingua(BasePassageCompressor):
	def __init__(
		self, project_dir: str, model_name: str = "NousResearch/Llama-2-7b-hf", **kwargs
	):
		try:
			from llmlingua import PromptCompressor
		except ImportError:
			raise ImportError(
				"LongLLMLingua is not installed. Please install it by running `pip install llmlingua`."
			)

		super().__init__(project_dir)
		model_init_params = pop_params(PromptCompressor.__init__, kwargs)
		self.llm_lingua = PromptCompressor(model_name=model_name, **model_init_params)

	def __del__(self):
		del self.llm_lingua
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, retrieved_contents = self.cast_to_run(previous_result)
		results = self._pure(queries, retrieved_contents, **kwargs)
		return list(map(lambda x: [x], results))

	def _pure(
		self,
		queries: List[str],
		contents: List[List[str]],
		instructions: Optional[str] = None,
		target_token: int = 300,
		**kwargs,
	) -> List[str]:
		"""
		Compresses the retrieved texts using LongLLMLingua.
		For more information, visit https://github.com/microsoft/LLMLingua.

		:param queries: The queries for retrieved passages.
		:param contents: The contents of retrieved passages.
		:param model_name: The model name to use for compression.
		    The default is "NousResearch/Llama-2-7b-hf".
		:param instructions: The instructions for compression.
		    Default is None. When it is None, it will use default instructions.
		:param target_token: The target token for compression.
		    Default is 300.
		:param kwargs: Additional keyword arguments.
		:return: The list of compressed texts.
		"""
		if instructions is None:
			instructions = "Given the context, please answer the final question"
		results = [
			llmlingua_pure(
				query, contents_, self.llm_lingua, instructions, target_token, **kwargs
			)
			for query, contents_ in zip(queries, contents)
		]

		return results


def llmlingua_pure(
	query: str,
	contents: List[str],
	llm_lingua,
	instructions: str,
	target_token: int = 300,
	**kwargs,
) -> str:
	"""
	Return the compressed text.

	:param query: The query for retrieved passages.
	:param contents: The contents of retrieved passages.
	:param llm_lingua: The llm instance, that will be used to compress.
	:param instructions: The instructions for compression.
	:param target_token: The target token for compression.
	    Default is 300.
	:param kwargs: Additional keyword arguments.
	:return: The compressed text.
	"""
	try:
		from llmlingua import PromptCompressor
	except ImportError:
		raise ImportError(
			"LongLLMLingua is not installed. Please install it by running `pip install llmlingua`."
		)
	# split by "\n\n" (recommended by LongLLMLingua authors)
	new_context_texts = [c for context in contents for c in context.split("\n\n")]
	compress_prompt_params = pop_params(PromptCompressor.compress_prompt, kwargs)
	compressed_prompt = llm_lingua.compress_prompt(
		new_context_texts,
		question=query,
		instruction=instructions,
		rank_method="longllmlingua",
		target_token=target_token,
		**compress_prompt_params,
	)
	compressed_prompt_txt = compressed_prompt["compressed_prompt"]

	# separate out the question and instruction
	result = "\n\n".join(compressed_prompt_txt.split("\n\n")[1:-1])

	return result



from .pass_passage_filter import PassPassageFilter
from .percentile_cutoff import PercentileCutoff
from .recency import RecencyFilter
from .similarity_percentile_cutoff import SimilarityPercentileCutoff
from .similarity_threshold_cutoff import SimilarityThresholdCutoff
from .threshold_cutoff import ThresholdCutoff



import itertools
from typing import Dict, List

from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.llms.openai.utils import to_openai_message_dicts
from openai import AsyncClient
from pydantic import BaseModel

from autorag.data.qa.query.prompt import QUERY_GEN_PROMPT


class Response(BaseModel):
	query: str


# Single hop QA generation OpenAI
async def query_gen_openai_base(
	row: Dict,
	client: AsyncClient,
	messages: List[ChatMessage],
	model_name: str = "gpt-4o-2024-08-06",
):
	context = list(itertools.chain.from_iterable(row["retrieval_gt_contents"]))
	context_str = "Text:\n" + "\n".join(
		[f"{i + 1}. {c}" for i, c in enumerate(context)]
	)
	user_prompt = f"{context_str}\n\nGenerated Question from the Text:\n"
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=to_openai_message_dicts(messages),
		response_format=Response,
	)
	row["query"] = completion.choices[0].message.parsed.query
	return row


async def factoid_query_gen(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	return await query_gen_openai_base(
		row, client, QUERY_GEN_PROMPT["factoid_single_hop"][lang], model_name
	)


async def concept_completion_query_gen(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	return await query_gen_openai_base(
		row, client, QUERY_GEN_PROMPT["factoid_single_hop"][lang], model_name
	)


class TwoHopIncrementalResponse(BaseModel):
	answer: str
	one_hop_question: str
	two_hop_question: str


async def two_hop_incremental(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	"""
	Create a two-hop question using incremental prompt.
	Incremental prompt is more effective to create multi-hop question.
	The input retrieval_gt has to include more than one passage.

	:return: The two-hop question using openai incremental prompt
	"""
	messages = QUERY_GEN_PROMPT["two_hop_incremental"][lang]
	passages = row["retrieval_gt_contents"]
	assert (
		len(passages) >= 2
	), "You have to sample more than two passages for making two-hop questions."
	context_str = f"Document 1: {passages[0][0]}\nDocument 2: {passages[1][0]}"
	user_prompt = f"{context_str}\n\nGenerated two-hop Question from two Documents:\n"
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=to_openai_message_dicts(messages),
		response_format=TwoHopIncrementalResponse,
	)
	row["query"] = completion.choices[0].message.parsed.two_hop_question
	return row



import os
import pathlib

import click
from datasets import load_dataset


@click.command()
@click.option(
	"--save_path",
	type=str,
	default=pathlib.PurePath(__file__).parent,
	help="Path to save sample msmarco dataset.",
)
def load_msmarco_dataset(save_path):
	# set file path
	file_path = "MarkrAI/msmarco_sample_autorag"

	# load dataset
	corpus_dataset = load_dataset(file_path, "corpus")["train"].to_pandas()
	qa_train_dataset = load_dataset(file_path, "qa")["train"].to_pandas()
	qa_test_dataset = load_dataset(file_path, "qa")["test"].to_pandas()

	# save corpus data
	if os.path.exists(os.path.join(save_path, "corpus.parquet")) is True:
		raise ValueError("corpus.parquet already exists")
	if os.path.exists(os.path.join(save_path, "qa.parquet")) is True:
		raise ValueError("qa.parquet already exists")
	corpus_dataset.to_parquet(os.path.join(save_path, "corpus.parquet"), index=False)
	qa_train_dataset.to_parquet(
		os.path.join(save_path, "qa_train.parquet"), index=False
	)
	qa_test_dataset.to_parquet(os.path.join(save_path, "qa_test.parquet"), index=False)


if __name__ == "__main__":
	load_msmarco_dataset()



from pathlib import Path
from typing import Tuple, List, Union

import numpy as np
import pandas as pd

from autorag.nodes.hybridretrieval.base import HybridRetrieval
from autorag.nodes.hybridretrieval.run import evaluate_retrieval_node
from autorag.strategy import select_best


def normalize_mm(scores: List[str], fixed_min_value: float = 0):
	arr = np.array(scores)
	max_value = np.max(arr)
	min_value = np.min(arr)
	norm_score = (arr - min_value) / (max_value - min_value)
	return norm_score


def normalize_tmm(scores: List[str], fixed_min_value: float):
	arr = np.array(scores)
	max_value = np.max(arr)
	norm_score = (arr - fixed_min_value) / (max_value - fixed_min_value)
	return norm_score


def normalize_z(scores: List[str], fixed_min_value: float = 0):
	arr = np.array(scores)
	mean_value = np.mean(arr)
	std_value = np.std(arr)
	norm_score = (arr - mean_value) / std_value
	return norm_score


def normalize_dbsf(scores: List[str], fixed_min_value: float = 0):
	arr = np.array(scores)
	mean_value = np.mean(arr)
	std_value = np.std(arr)
	min_value = mean_value - 3 * std_value
	max_value = mean_value + 3 * std_value
	norm_score = (arr - min_value) / (max_value - min_value)
	return norm_score


normalize_method_dict = {
	"mm": normalize_mm,
	"tmm": normalize_tmm,
	"z": normalize_z,
	"dbsf": normalize_dbsf,
}


class HybridCC(HybridRetrieval):
	def _pure(
		self,
		info: dict,
		top_k: int,
		weight: float,
		normalize_method: str = "mm",
		semantic_theoretical_min_value: float = -1.0,
		lexical_theoretical_min_value: float = 0.0,
	):
		return hybrid_cc(
			(info["retrieved_ids_semantic"], info["retrieved_ids_lexical"]),
			(info["retrieve_scores_semantic"], info["retrieve_scores_lexical"]),
			top_k,
			weight,
			normalize_method,
			semantic_theoretical_min_value,
			lexical_theoretical_min_value,
		)

	@classmethod
	def run_evaluator(
		cls,
		project_dir: Union[str, Path],
		previous_result: pd.DataFrame,
		*args,
		**kwargs,
	):
		assert "strategy" in kwargs, "You must specify the strategy to use."
		assert (
			"input_metrics" in kwargs
		), "You must specify the input metrics to use, which is list of MetricInput."
		strategies = kwargs.pop("strategy")
		input_metrics = kwargs.pop("input_metrics")
		weight_range = kwargs.pop("weight_range", (0.0, 1.0))
		test_weight_size = kwargs.pop("test_weight_size", 101)
		weight_candidates = np.linspace(
			weight_range[0], weight_range[1], test_weight_size
		).tolist()

		result_list = []
		instance = cls(project_dir, *args, **kwargs)
		for weight_value in weight_candidates:
			result_df = instance.pure(previous_result, weight=weight_value, **kwargs)
			result_list.append(result_df)

		if strategies.get("metrics") is None:
			raise ValueError("You must at least one metrics for retrieval evaluation.")
		result_list = list(
			map(
				lambda x: evaluate_retrieval_node(
					x,
					input_metrics,
					strategies.get("metrics"),
				),
				result_list,
			)
		)

		# select best result
		best_result_df, best_weight = select_best(
			result_list,
			strategies.get("metrics"),
			metadatas=weight_candidates,
			strategy_name=strategies.get("strategy", "normalize_mean"),
		)
		return {
			"best_result": best_result_df,
			"best_weight": best_weight,
		}


def hybrid_cc(
	ids: Tuple,
	scores: Tuple,
	top_k: int,
	weight: float,
	normalize_method: str = "mm",
	semantic_theoretical_min_value: float = -1.0,
	lexical_theoretical_min_value: float = 0.0,
) -> Tuple[List[List[str]], List[List[float]]]:
	"""
	Hybrid CC function.
	CC (convex combination) is a method to fuse lexical and semantic retrieval results.
	It is a method that first normalizes the scores of each retrieval result,
	and then combines them with the given weights.
	It is uniquer than other retrieval modules, because it does not really execute retrieval,
	but just fuse the results of other retrieval functions.
	So you have to run more than two retrieval modules before running this function.
	And collect ids and scores result from each retrieval module.
	Make it as tuple and input it to this function.

	:param ids: The tuple of ids that you want to fuse.
	    The length of this must be the same as the length of scores.
	    The semantic retrieval ids must be the first index.
	:param scores: The retrieve scores that you want to fuse.
	    The length of this must be the same as the length of ids.
	    The semantic retrieval scores must be the first index.
	:param top_k: The number of passages to be retrieved.
	:param normalize_method: The normalization method to use.
	  There are some normalization method that you can use at the hybrid cc method.
	  AutoRAG support following.
	    - `mm`: Min-max scaling
	    - `tmm`: Theoretical min-max scaling
	    - `z`: z-score normalization
	    - `dbsf`: 3-sigma normalization
	:param weight: The weight value. If the weight is 1.0, it means the
	  weight to the semantic module will be 1.0 and weight to the lexical module will be 0.0.
	:param semantic_theoretical_min_value: This value used by `tmm` normalization method. You can set the
	    theoretical minimum value by yourself. Default is -1.
	:param lexical_theoretical_min_value: This value used by `tmm` normalization method. You can set the
	    theoretical minimum value by yourself. Default is 0.
	:return: The tuple of ids and fused scores that fused by CC. Plus, the third element is selected weight value.
	"""
	assert len(ids) == len(scores), "The length of ids and scores must be the same."
	assert len(ids) > 1, "You must input more than one retrieval results."
	assert top_k > 0, "top_k must be greater than 0."
	assert weight >= 0, "The weight must be greater than 0."
	assert weight <= 1, "The weight must be less than 1."

	df = pd.DataFrame(
		{
			"semantic_ids": ids[0],
			"lexical_ids": ids[1],
			"semantic_score": scores[0],
			"lexical_score": scores[1],
		}
	)

	def cc_pure_apply(row):
		return fuse_per_query(
			row["semantic_ids"],
			row["lexical_ids"],
			row["semantic_score"],
			row["lexical_score"],
			normalize_method=normalize_method,
			weight=weight,
			top_k=top_k,
			semantic_theoretical_min_value=semantic_theoretical_min_value,
			lexical_theoretical_min_value=lexical_theoretical_min_value,
		)

	# fixed weight
	df[["cc_id", "cc_score"]] = df.apply(
		lambda row: cc_pure_apply(row), axis=1, result_type="expand"
	)
	return df["cc_id"].tolist(), df["cc_score"].tolist()


def fuse_per_query(
	semantic_ids: List[str],
	lexical_ids: List[str],
	semantic_scores: List[float],
	lexical_scores: List[float],
	normalize_method: str,
	weight: float,
	top_k: int,
	semantic_theoretical_min_value: float,
	lexical_theoretical_min_value: float,
):
	normalize_func = normalize_method_dict[normalize_method]
	norm_semantic_scores = normalize_func(
		semantic_scores, semantic_theoretical_min_value
	)
	norm_lexical_scores = normalize_func(lexical_scores, lexical_theoretical_min_value)
	ids = [semantic_ids, lexical_ids]
	scores = [norm_semantic_scores, norm_lexical_scores]
	df = pd.concat(
		[pd.Series(dict(zip(_id, score))) for _id, score in zip(ids, scores)], axis=1
	)
	df.columns = ["semantic", "lexical"]
	df = df.fillna(0)
	df["weighted_sum"] = df.mul((weight, 1.0 - weight)).sum(axis=1)
	df = df.sort_values(by="weighted_sum", ascending=False)
	return df.index.tolist()[:top_k], df["weighted_sum"][:top_k].tolist()



import uuid
from typing import List, Optional

import pandas as pd
from langchain_core.documents import Document

from autorag.data.utils.util import add_essential_metadata
from autorag.utils.util import save_parquet_safe


def langchain_documents_to_parquet(
	langchain_documents: List[Document],
	output_filepath: Optional[str] = None,
	upsert: bool = False,
) -> pd.DataFrame:
	"""
	Langchain documents to corpus dataframe.
	Corpus dataframe will be saved to filepath(file_dir/filename) if given.
	Return corpus dataframe whether the filepath is given.
	You can use this method to create corpus.parquet after load and chunk using Llama Index.

	:param langchain_documents: List of langchain documents.
	:param output_filepath: Optional filepath to save the parquet file.
	    If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
	    File directory must exist. File extension must be .parquet
	:param upsert: If true, the function will overwrite the existing file if it exists.
	    Default is False.
	:return: Corpus data as pd.DataFrame
	"""

	corpus_df = pd.DataFrame(
		list(
			map(
				lambda doc: {
					"doc_id": str(uuid.uuid4()),
					"contents": doc.page_content,
					"metadata": add_essential_metadata(doc.metadata),
				},
				langchain_documents,
			)
		)
	)

	if output_filepath is not None:
		save_parquet_safe(corpus_df, output_filepath, upsert=upsert)

	return corpus_df



import pandas as pd

from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.utils import result_to_dataframe


class PassPassageFilter(BasePassageFilter):
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		_, contents, scores, ids = self.cast_to_run(previous_result)
		return contents, ids, scores

	def _pure(self, *args, **kwargs):
		pass



from typing import List, Tuple

import numpy as np
import pandas as pd

from autorag.embedding.base import EmbeddingModel
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.utils.util import (
	embedding_query_content,
	empty_cuda_cache,
	result_to_dataframe,
	pop_params,
)


class SimilarityThresholdCutoff(BasePassageFilter):
	def __init__(self, project_dir: str, *args, **kwargs):
		"""
		Initialize the SimilarityThresholdCutoff module

		:param project_dir: The project directory to use for initializing the module
		:param embedding_model: The embedding model string to use for calculating similarity
		        Default is "openai" which is OpenAI text-embedding-ada-002 embedding model.
		"""
		super().__init__(project_dir, *args, **kwargs)
		embedding_model = kwargs.get("embedding_model", "openai")
		self.embedding_model = EmbeddingModel.load(embedding_model)()

	def __del__(self):
		del self.embedding_model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		kwargs = pop_params(self._pure, kwargs)
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		return self._pure(queries, contents, scores, ids, *args, **kwargs)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		threshold: float,
		batch: int = 128,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Re-calculate each content's similarity with the query and filter out the contents that are below the threshold.
		If all contents are filtered, keep the only one highest similarity content.
		This is a filter and does not override scores.
		The output of scores is not coming from query-content similarity.

		:param queries: The list of queries to use for filtering
		:param contents_list: The list of lists of contents to filter
		:param scores_list: The list of lists of scores retrieved
		:param ids_list: The list of lists of ids retrieved
		:param threshold: The threshold to cut off
		:param batch: The number of queries to be processed in a batch
		    Default is 128.
		:return: Tuple of lists containing the filtered contents, ids, and scores
		"""
		query_embeddings, content_embeddings = embedding_query_content(
			queries, contents_list, self.embedding_model, batch
		)

		remain_indices = list(
			map(
				lambda x: self.__row_pure(x[0], x[1], threshold),
				zip(query_embeddings, content_embeddings),
			)
		)

		remain_content_list = list(
			map(lambda c, idx: [c[i] for i in idx], contents_list, remain_indices)
		)
		remain_scores_list = list(
			map(lambda s, idx: [s[i] for i in idx], scores_list, remain_indices)
		)
		remain_ids_list = list(
			map(lambda _id, idx: [_id[i] for i in idx], ids_list, remain_indices)
		)
		return remain_content_list, remain_ids_list, remain_scores_list

	@staticmethod
	def __row_pure(
		query_embedding: str, content_embeddings: List[List[float]], threshold: float
	) -> List[int]:
		"""
		Return indices that have to remain.
		Return at least one index if there is nothing to remain.

		:param query_embedding: Query embedding
		:param content_embeddings: Each content embedding
		:param threshold: The threshold to cut off
		:return: Indices to remain at the contents
		"""

		similarities = np.array(
			list(
				map(
					lambda x: calculate_cosine_similarity(query_embedding, x),
					content_embeddings,
				)
			)
		)
		result = np.where(similarities >= threshold)[0].tolist()
		if len(result) > 0:
			return result
		return [np.argmax(similarities)]



from .preprocess import (
	validate_qa_dataset,
	validate_corpus_dataset,
	cast_qa_dataset,
	cast_corpus_dataset,
	validate_qa_from_corpus_dataset,
)
from .util import fetch_contents, result_to_dataframe, sort_by_scores



from dataclasses import dataclass
from typing import Optional, List, Dict, Callable, Any, Union

import numpy as np
import pandas as pd


@dataclass
class MetricInput:
	query: Optional[str] = None
	queries: Optional[List[str]] = None
	retrieval_gt_contents: Optional[List[List[str]]] = None
	retrieved_contents: Optional[List[str]] = None
	retrieval_gt: Optional[List[List[str]]] = None
	retrieved_ids: Optional[List[str]] = None
	prompt: Optional[str] = None
	generated_texts: Optional[str] = None
	generation_gt: Optional[List[str]] = None
	generated_log_probs: Optional[List[float]] = None

	def is_fields_notnone(self, fields_to_check: List[str]) -> bool:
		for field in fields_to_check:
			actual_value = getattr(self, field)

			if actual_value is None:
				return False

			try:
				if not type_checks.get(type(actual_value), lambda _: False)(
					actual_value
				):
					return False
			except Exception:
				return False

		return True

	@classmethod
	def from_dataframe(cls, qa_data: pd.DataFrame) -> List["MetricInput"]:
		"""
		Convert a pandas DataFrame into a list of MetricInput instances.
		qa_data: pd.DataFrame: qa_data DataFrame containing metric data.

		:returns: List[MetricInput]: List of MetricInput objects created from DataFrame rows.
		"""
		instances = []

		for _, row in qa_data.iterrows():
			instance = cls()

			for attr_name in cls.__annotations__:
				if attr_name in row:
					value = row[attr_name]

					if isinstance(value, str):
						setattr(
							instance,
							attr_name,
							value.strip() if value.strip() != "" else None,
						)
					elif isinstance(value, list):
						setattr(instance, attr_name, value if len(value) > 0 else None)
					else:
						setattr(instance, attr_name, value)

			instances.append(instance)

		return instances

	@staticmethod
	def _check_list(lst_or_arr: Union[List[Any], np.ndarray]) -> bool:
		if isinstance(lst_or_arr, np.ndarray):
			lst_or_arr = lst_or_arr.flatten().tolist()

		if len(lst_or_arr) == 0:
			return False

		for item in lst_or_arr:
			if item is None:
				return False

			item_type = type(item)

			if item_type in type_checks:
				if not type_checks[item_type](item):
					return False
			else:
				return False

		return True


type_checks: Dict[type, Callable[[Any], bool]] = {
	str: lambda x: len(x.strip()) > 0,
	list: MetricInput._check_list,
	np.ndarray: MetricInput._check_list,
	int: lambda _: True,
	float: lambda _: True,
}



import functools
from typing import List, Callable, Any, Tuple

import pandas as pd

from autorag.evaluation.metric import (
	retrieval_token_f1,
	retrieval_token_precision,
	retrieval_token_recall,
)
from autorag.schema.metricinput import MetricInput


def evaluate_retrieval_contents(metric_inputs: List[MetricInput], metrics: List[str]):
	def decorator_evaluate_retrieval_contents(
		func: Callable[
			[Any], Tuple[List[List[str]], List[List[str]], List[List[float]]]
		],
	):
		"""
		Decorator for evaluating retrieval contents.
		You can use this decorator to any method that returns (contents, scores, ids),
		which is the output of conventional retrieval modules.

		:param func: Must return (contents, scores, ids)
		:return: pd.DataFrame, which is the evaluation result and function result.
		"""

		@functools.wraps(func)
		def wrapper(*args, **kwargs) -> pd.DataFrame:
			contents, pred_ids, scores = func(*args, **kwargs)
			metric_funcs = {
				retrieval_token_recall.__name__: retrieval_token_recall,
				retrieval_token_precision.__name__: retrieval_token_precision,
				retrieval_token_f1.__name__: retrieval_token_f1,
			}
			for metric_input, content in zip(metric_inputs, contents):
				metric_input.retrieved_contents = content

			metrics_scores = {}
			for metric in metrics:
				if metric not in metric_funcs:
					raise ValueError(
						f"metric {metric} is not in supported metrics: {metric_funcs.keys()}"
					)
				else:
					metric_func = metric_funcs[metric]
					# Extract each required field from all payloads
					metric_scores = metric_func(metric_inputs=metric_inputs)
					metrics_scores[metric] = metric_scores

			metric_result_df = pd.DataFrame(metrics_scores)
			execution_result_df = pd.DataFrame(
				{
					"retrieved_contents": contents,
					"retrieved_ids": pred_ids,
					"retrieve_scores": scores,
				}
			)
			result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
			return result_df

		return wrapper

	return decorator_evaluate_retrieval_contents



import os
from typing import List, Tuple

import aiohttp
import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import get_event_loop, process_batch, result_to_dataframe

JINA_API_URL = "https://api.jina.ai/v1/rerank"


class JinaReranker(BasePassageReranker):
	def __init__(self, project_dir: str, api_key: str = None, *args, **kwargs):
		"""
		Initialize Jina rerank node.

		:param project_dir: The project directory path.
		:param api_key: The API key for Jina rerank.
		You can set it in the environment variable JINAAI_API_KEY.
		Or, you can directly set it on the config YAML file using this parameter.
		Default is env variable "JINAAI_API_KEY".
		:param kwargs: Extra arguments that are not affected
		"""
		super().__init__(project_dir)
		if api_key is None:
			api_key = os.getenv("JINAAI_API_KEY", None)
			if api_key is None:
				raise ValueError(
					"API key is not provided."
					"You can set it as an argument or as an environment variable 'JINAAI_API_KEY'"
				)
		self.session = aiohttp.ClientSession(loop=get_event_loop())
		self.session.headers.update(
			{"Authorization": f"Bearer {api_key}", "Accept-Encoding": "identity"}
		)

	def __del__(self):
		self.session.close()
		del self.session
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 8)
		model = kwargs.pop("model", "jina-reranker-v1-base-en")
		return self._pure(queries, contents, ids, top_k, model, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		model: str = "jina-reranker-v1-base-en",
		batch: int = 8,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with Jina rerank models.
		You can get the API key from https://jina.ai/reranker and set it in the environment variable JINAAI_API_KEY.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param model: The model name for Cohere rerank.
		    You can choose between "jina-reranker-v1-base-en" and "jina-colbert-v1-en".
		    Default is "jina-reranker-v1-base-en".
		:param batch: The number of queries to be processed in a batch
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		tasks = [
			jina_reranker_pure(
				self.session, query, contents, ids, top_k=top_k, model=model
			)
			for query, contents, ids in zip(queries, contents_list, ids_list)
		]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch))

		content_result, id_result, score_result = zip(*results)

		return list(content_result), list(id_result), list(score_result)


async def jina_reranker_pure(
	session,
	query: str,
	contents: List[str],
	ids: List[str],
	top_k: int,
	model: str = "jina-reranker-v1-base-en",
) -> Tuple[List[str], List[str], List[float]]:
	async with session.post(
		JINA_API_URL,
		json={
			"query": query,
			"documents": contents,
			"model": model,
			"top_n": top_k,
		},
	) as resp:
		resp_json = await resp.json()
		if "results" not in resp_json:
			raise RuntimeError(f"Invalid response from Jina API: {resp_json['detail']}")

		results = resp_json["results"]
		indices = list(map(lambda x: x["index"], results))
		score_result = list(map(lambda x: x["relevance_score"], results))
		id_result = list(map(lambda x: ids[x], indices))
		content_result = list(map(lambda x: contents[x], indices))

		return content_result, id_result, score_result



from .generation import (
	bleu,
	meteor,
	rouge,
	sem_score,
	g_eval,
	bert_score,
	deepeval_faithfulness,
)
from .retrieval import (
	retrieval_f1,
	retrieval_recall,
	retrieval_precision,
	retrieval_mrr,
	retrieval_ndcg,
	retrieval_map,
)
from .retrieval_contents import (
	retrieval_token_f1,
	retrieval_token_precision,
	retrieval_token_recall,
)



from typing import List

import pandas as pd

from autorag.nodes.passagecompressor.base import BasePassageCompressor
from autorag.utils import result_to_dataframe


class PassCompressor(BasePassageCompressor):
	@result_to_dataframe(["retrieved_contents"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		_, contents = self.cast_to_run(previous_result)
		return self._pure(contents)

	def _pure(self, contents: List[List[str]]):
		return contents



import logging
import os
from datetime import datetime, date
from pathlib import Path
from typing import List, Tuple, Union

import pandas as pd

from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.utils import fetch_contents, result_to_dataframe

logger = logging.getLogger("AutoRAG")


class RecencyFilter(BasePassageFilter):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		super().__init__(project_dir, *args, **kwargs)
		self.corpus_df = pd.read_parquet(
			os.path.join(project_dir, "data", "corpus.parquet"), engine="pyarrow"
		)

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		_, contents, scores, ids = self.cast_to_run(previous_result, *args, **kwargs)
		metadatas = fetch_contents(self.corpus_df, ids, column_name="metadata")
		times = [
			[time["last_modified_datetime"] for time in time_list]
			for time_list in metadatas
		]
		return self._pure(contents, scores, ids, times, *args, **kwargs)

	def _pure(
		self,
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		time_list: List[List[datetime]],
		threshold_datetime: Union[datetime, date],
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Filter out the contents that are below the threshold datetime.
		If all contents are filtered, keep the only one recency content.
		If the threshold date format is incorrect, return the original contents.

		:param contents_list: The list of lists of contents to filter
		:param scores_list: The list of lists of scores retrieved
		:param ids_list: The list of lists of ids retrieved
		:param time_list: The list of lists of datetime retrieved
		:param threshold_datetime: The threshold to cut off.
			In recency filter, you have to use the datetime.datetime object or datetime.date object.
			All you need to do is to set the date at your YAML file.
			For example, you can write "2010-09-09 3:45:06" or "2010-09-09" in the YAML file.
		:return: Tuple of lists containing the filtered contents, ids, and scores
		"""
		if not (
			isinstance(threshold_datetime, datetime)
			or isinstance(threshold_datetime, date)
		):
			raise ValueError(
				f"Threshold should be a datetime object, but got {type(threshold_datetime)}"
			)

		if not isinstance(threshold_datetime, datetime):
			threshold_datetime = datetime.combine(
				threshold_datetime, datetime.min.time()
			)

		time_list = [
			list(
				map(
					lambda t: datetime.combine(t, datetime.min.time())
					if not isinstance(t, datetime)
					else t,
					time,
				)
			)
			for time in time_list
		]

		def sort_row(contents, scores, ids, time, _datetime_threshold):
			combined = list(zip(contents, scores, ids, time))
			combined_filtered = [
				item for item in combined if item[3] >= _datetime_threshold
			]

			if combined_filtered:
				remain_contents, remain_scores, remain_ids, _ = zip(*combined_filtered)
			else:
				combined.sort(key=lambda x: x[3], reverse=True)
				remain_contents, remain_scores, remain_ids, _ = zip(*combined[:1])

			return list(remain_contents), list(remain_ids), list(remain_scores)

		remain_contents_list, remain_ids_list, remain_scores_list = zip(
			*map(
				sort_row,
				contents_list,
				scores_list,
				ids_list,
				time_list,
				[threshold_datetime] * len(contents_list),
			)
		)

		return remain_contents_list, remain_ids_list, remain_scores_list



from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.utils.util import convert_inputs_to_list, result_to_dataframe


class ThresholdCutoff(BasePassageFilter):
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		_, contents, scores, ids = self.cast_to_run(previous_result)
		return self._pure(contents, scores, ids, *args, **kwargs)

	def _pure(
		self,
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		threshold: float,
		reverse: bool = False,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Filters the contents, scores, and ids based on a previous result's score.
		Keeps at least one item per query if all scores are below the threshold.

		:param contents_list: List of content strings for each query.
		:param scores_list: List of scores for each content.
		:param ids_list: List of ids for each content.
		:param threshold: The minimum score to keep an item.
		:param reverse: If True, the lower the score, the better.
		    Default is False.
		:return: Filtered lists of contents, ids, and scores.
		"""
		remain_indices = list(
			map(lambda x: self.__row_pure(x, threshold, reverse), scores_list)
		)

		remain_content_list = list(
			map(lambda c, idx: [c[i] for i in idx], contents_list, remain_indices)
		)
		remain_scores_list = list(
			map(lambda s, idx: [s[i] for i in idx], scores_list, remain_indices)
		)
		remain_ids_list = list(
			map(lambda _id, idx: [_id[i] for i in idx], ids_list, remain_indices)
		)

		return remain_content_list, remain_ids_list, remain_scores_list

	@convert_inputs_to_list
	def __row_pure(
		self, scores_list: List[float], threshold: float, reverse: bool = False
	) -> List[int]:
		"""
		Return indices that have to remain.
		Return at least one index if there is nothing to remain.

		:param scores_list: Each score
		:param threshold: The threshold to cut off
		:param reverse: If True, the lower the score, the better
			Default is False.
		:return: Indices to remain at the contents
		"""
		assert isinstance(scores_list, list), "scores_list must be a list."

		if reverse:
			remain_indices = [
				i for i, score in enumerate(scores_list) if score <= threshold
			]
			default_index = scores_list.index(min(scores_list))
		else:
			remain_indices = [
				i for i, score in enumerate(scores_list) if score >= threshold
			]
			default_index = scores_list.index(max(scores_list))

		return remain_indices if remain_indices else [default_index]



import os
from datetime import datetime
from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils import result_to_dataframe, fetch_contents


class TimeReranker(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir, *args, **kwargs)
		self.corpus_df = pd.read_parquet(
			os.path.join(project_dir, "data", "corpus.parquet"), engine="pyarrow"
		)

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		_, contents, scores, ids = self.cast_to_run(previous_result)
		metadatas = fetch_contents(self.corpus_df, ids, column_name="metadata")
		times = [
			[time["last_modified_datetime"] for time in time_list]
			for time_list in metadatas
		]
		top_k = kwargs.pop("top_k")
		return self._pure(contents, scores, ids, top_k, times)

	def _pure(
		self,
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		top_k: int,
		time_list: List[List[datetime]],
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank the passages based on merely the datetime of the passage.
		It uses 'last_modified_datetime' key in the corpus metadata,
		so the metadata should be in the format of {'last_modified_datetime': datetime.datetime} at the corpus data file.

		:param contents_list: The list of lists of contents
		:param scores_list: The list of lists of scores from the initial ranking
		:param ids_list: The list of lists of ids
		:param top_k: The number of passages to be retrieved after reranking
		:param time_list: The metadata list of lists of datetime.datetime
			It automatically extracts the 'last_modified_datetime' key from the metadata in the corpus data.
		:return: The reranked contents, ids, and scores
		"""

		def sort_row(contents, scores, ids, time, top_k):
			combined = list(zip(contents, scores, ids, time))
			combined.sort(key=lambda x: x[3], reverse=True)
			sorted_contents, sorted_scores, sorted_ids, _ = zip(*combined)
			return (
				list(sorted_contents)[:top_k],
				list(sorted_scores)[:top_k],
				list(sorted_ids)[:top_k],
			)

		reranked_contents, reranked_scores, reranked_ids = zip(
			*map(
				sort_row,
				contents_list,
				scores_list,
				ids_list,
				time_list,
				[top_k] * len(contents_list),
			)
		)

		return list(reranked_contents), list(reranked_ids), list(reranked_scores)
